/* GDCHART 2  GDCHART.C
 *
 ************************************************************************
 * This version of the GDChart library was based on GDCHART 0.94b
 * (12 Nov. 1998) by Bruce Verderaime (brv@fred.net)
 *
 * 2000-11-20 - Merged in changes from GDCHART 0.10.1dev
 *
 * Here is the original copyright notice:
 *
 +-----------------------------------------------------------------------
 | GDChart is free for use in your applications and for chart generation.
 | YOU MAY NOT re-distribute or represent the code as your own.
 | Any re-distributions of the code MUST reference the author, and include
 | any and all original documentaion.
 +-----------------------------------------------------------------------
 *
 * $Log: gdchart.c,v $
 * Revision 1.13  2003/01/27 22:55:20  daniel
 * Fixed compile error (vars declared after first statement in {} block)
 *
 * Revision 1.12  2002/11/15 18:58:34  julien
 * implement Bar Labels
 *
 * Revision 1.11  2002/02/11 19:51:10  julien
 * Add Combo_Line_HLC and 3DCombo_Line_HLC type
 *
 * Revision 1.10  2002/01/21 21:30:05  julien
 * Valid if gif, png, jpeg or wbmp exist
 *
 * Revision 1.9  2002/01/14 17:35:50  julien
 * Added ImageType and JPEGQuality parameters to create GIF, PNG and JPEG
 *
 * Revision 1.8  2002/01/11 15:51:16  julien
 * Add Combo_HLC_Line and 3DCombo_HLC_Line type
 *
 * Revision 1.7  2000/11/24 18:51:07  daniel
 * Fixed some problems with H2BAR labels and grid
 *
 * Revision 1.6  2000/11/22 21:48:17  daniel
 * Fixed test on vhighest/lowest for yaxis2 labels when requested_ymin2/max2
 * is used
 *
 * Revision 1.5  2000/11/20 19:58:13  daniel
 * Integrate changes between v0.94b and v0.10.1dev
 *
 * Revision 1.4  2000/11/16 21:30:13  daniel
 * With combo charts, if yaxis2==FALSE then use same Y min/max for both sets
 *
 * Revision 1.3  1999/08/03 15:52:01  daniel
 * Adjust placement of left/right labels in H2BAR charts
 *
 * Revision 1.2  1999/08/03 15:17:10  daniel
 * Fix some problems with BAR, HBAR, H2BAR charts and log scale
 *
 */

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>

#ifdef _WIN32
#  include <stdlib.h>
#  pragma warning( once : 4244 ) /* Issue warning 4244 only once. */
#else
#  include <values.h>
#endif

#include "gdc.h"    /* Note: External programs should include gdchart.h only */


#define HIGHSET     0
#define LOWSET      1
#define CLOSESET    2

#define H2_LEFTSET     0  /* for H2BAR types only */
#define H2_RIGHTSET    1  /* RIGHTSET used for HBAR types as well */

/* scaled translation onto graph */
#define PX( x )     (int)( xorig + (setno*xdepth_3D) + (x)*xscl )
#define PY( y )     (int)( yorig - (setno*ydepth_3D) + (y)*yscl )
#define PV( y )     (int)( vyorig - (setno*ydepth_3D) + (y)*vyscl )

/* PXH2() for H2BAR types, side= -1 for left side, and 1 for right side */
#define PXH2( side, x )     (int)( false_xorig + (setno*xdepth_3D) + ((x)-lowest)*(side)*xscl )


#define SET_RECT( gdp, x1, x2, y1, y2 ) gdp[0].x = gdp[3].x = x1,   \
                                        gdp[0].y = gdp[1].y = y1,   \
                                        gdp[1].x = gdp[2].x = x2,   \
                                        gdp[2].y = gdp[3].y = y2

#ifdef THUMB_VALS
/**********************************************************************
 *                       smallarrow()
 **********************************************************************/
/* -------------------------------------------------------------------
 * draw an arrow at (x,y)-upper left in arrwclr to the size of SmallFont
 * could, with just a little difficulty, be made to accept a font size
 * ------------------------------------------------------------------- */
static void
smallarrow( gdImagePtr  im,
            int         x,
            int         y,
            char        up,
            int         arrwclr )
{
    gdImageLine( im, x+2, y, x+2, y+GDC_fontc[GDC_SMALL].h, arrwclr );
    gdImageLine( im, x+3, y, x+3, y+GDC_fontc[GDC_SMALL].h, arrwclr );
    if( up )                                              /*   oo   */
        {                                                 /*  uoou  */
        gdImageSetPixel( im, x,   y+2, arrwclr );         /* uuoouu */
        gdImageSetPixel( im, x+1, y+2, arrwclr );         /*   oo   */
        gdImageSetPixel( im, x+4, y+2, arrwclr );         /*   oo   */
        gdImageSetPixel( im, x+5, y+2, arrwclr );         /*   oo   */
        gdImageSetPixel( im, x+1, y+1, arrwclr );         /*   oo   */
        gdImageSetPixel( im, x+4, y+1, arrwclr );         /*   oo   */
        }                                                 /*   oo   */
    else                                                  /* ddoodd */
        {                                                 /*  dood  */
                                                          /*   oo   */
        gdImageSetPixel( im, x,   y+(GDC_fontc[GDC_SMALL].h-2), arrwclr );  
        gdImageSetPixel( im, x+1, y+(GDC_fontc[GDC_SMALL].h-2), arrwclr );
        gdImageSetPixel( im, x+4, y+(GDC_fontc[GDC_SMALL].h-2), arrwclr );
        gdImageSetPixel( im, x+5, y+(GDC_fontc[GDC_SMALL].h-2), arrwclr );
        gdImageSetPixel( im, x+1, y+(GDC_fontc[GDC_SMALL].h-1), arrwclr );
        gdImageSetPixel( im, x+4, y+(GDC_fontc[GDC_SMALL].h-1), arrwclr );
        }
}
#endif


#define SET_3D_POLY( gdp, x1, x2, y1, y2, xoff, yoff )                      \
                                gdp[0].x  = x1,        gdp[0].y = y1,       \
                                gdp[1].x  = x1+(xoff), gdp[1].y = y1-yoff,  \
                                gdp[2].x  = x2+(xoff), gdp[2].y = y2-yoff,  \
                                gdp[3].x  = x2,        gdp[3].y = y2

/**********************************************************************
 *                       draw_3d_line()
 **********************************************************************/
/* ------------------------------------------------------------------------- */
/* vals in pixels */
/* ref is front plane */
/* allows for intersecting 3D lines      */
/*  (also used for single 3D lines >:-Q  */

struct YS { int y1; int y2; float slope; int lnclr; int shclr; };
static int qcmpr( const void *a, const void *b )
	{ if( ((struct YS*)a)->y2 < ((struct YS*)b)->y2 ) return 1;
	  if( ((struct YS*)a)->y2 > ((struct YS*)b)->y2 ) return -1;
	  return 0; }

static void
draw_3d_line( gdImagePtr    im,
              int           y0,
              int           x1,
              int           x2,
              int           y1[],
              int           y2[],
              int           xdepth,
              int           ydepth,
              int           num_sets,
              int           clr[],
              int           clrshd[] )
{
#define F(x,i)  (int)( (float)((x)-x1)*slope[i]+(float)y1[i] )
    float       depth_slope  = xdepth==0? MAXFLOAT: (float)ydepth/(float)xdepth;
    /*float    slope[num_sets]; */
    DEF_ARRAY(float, slope, num_sets);
    /*int           lnclr[num_sets],*/
    /*          shclr[num_sets]; */
    DEF_ARRAY(int,   lnclr, num_sets);
    DEF_ARRAY(int,   shclr, num_sets);
    int         i;
    int         x;
    gdPoint     poly[4];
    /*struct YS ypts[num_sets]; */
    DEF_ARRAY(struct YS, ypts, num_sets);

    ALLOC_ARRAY(float, slope, num_sets);
    ALLOC_ARRAY(int,   lnclr, num_sets);
    ALLOC_ARRAY(int,   shclr, num_sets);
    ALLOC_ARRAY(struct YS, ypts, num_sets);


    for( i=0; i<num_sets; ++i )
    {
        /* lnclr[i] = clr[i]; */
        /* shclr[i] = clrshd[i]; */
        slope[i] = x2==x1? MAXFLOAT: (float)(y2[i]-y1[i])/(float)(x2-x1);
    }

    for( x=x1+1; x<=x2; ++x )
    {
        for( i=0; i<num_sets; ++i )             /* load set of points */
        {
            ypts[i].y1    = F(x-1,i);
            ypts[i].y2    = F(x,i);
            ypts[i].lnclr = clr[i];
            ypts[i].shclr = clrshd[i];
            ypts[i].slope = slope[i];
        }                                       /* sorted "lowest" first */
        qsort( ypts, num_sets, sizeof(struct YS), qcmpr );
                                                /* put out in that order */
        for( i=0; i<num_sets; ++i )
        {                                       /* top */
            SET_3D_POLY( poly, x-1, x, ypts[i].y1, ypts[i].y2, 
                         xdepth, ydepth );
            gdImageFilledPolygon( im, poly, 4,  /* depth_slope ever < 0 ? */
                                  -ypts[i].slope>depth_slope? ypts[i].shclr: 
                                                              ypts[i].lnclr );
            if( x == x1+1 )                     /* edging */
                gdImageLine( im,
                             x-1, ypts[i].y2,
                             x-1+xdepth, ypts[i].y2-ydepth,
                             -ypts[i].slope<=depth_slope? ypts[i].shclr: 
                                                          ypts[i].lnclr );
        }
    }

    FREE_ARRAY(slope);
    FREE_ARRAY(lnclr);
    FREE_ARRAY(shclr);
    FREE_ARRAY(ypts);
}

/**********************************************************************
 *                       draw_3d_area()
 **********************************************************************/
/* ------------------------------------------------------------------------- */
/* vals in pixels */
/* ref is front plane */
static void
draw_3d_area( gdImagePtr        im,
             int                x1,
             int                x2,
             int                y0,         // drawn from 0
             int                y1,
             int                y2,
             int                xdepth,
             int                ydepth,
             int                clr,
             int                clrshd )
{

    gdPoint     poly[4];
    int         y_intercept = 0;        /* if xdepth || ydepth */

    if( xdepth || ydepth )
    {
        float   line_slope   = x2==x1?    MAXFLOAT: (float)-(y2-y1) / (float)(x2-x1);
        float   depth_slope  = xdepth==0? MAXFLOAT: (float)ydepth/(float)xdepth;

        y_intercept = (y1 > y0 && y2 < y0) ||           /* line crosses y0 */
                       (y1 < y0 && y2 > y0)?
                            (int)((1.0/ABS(line_slope))*(float)(ABS(y1-y0)))+x1:
                                                 0;     /* never */

                                          /* edging along y0 depth */
        gdImageLine( im, x1+xdepth, y0-ydepth, x2+xdepth, y0-ydepth, clrshd );

        SET_3D_POLY( poly, x1, x2, y1, y2, xdepth, ydepth );    /* top */
        gdImageFilledPolygon( im, poly, 4, 
                              line_slope>depth_slope? clrshd: clr );

        SET_3D_POLY( poly, x1, x2, y0, y0, xdepth, ydepth+1 ); /*along y axis*/
        gdImageFilledPolygon( im, poly, 4, clr );

        SET_3D_POLY( poly, x2, x2, y0, y2, xdepth, ydepth );    /* side */
        gdImageFilledPolygon( im, poly, 4, clrshd );

        if( y_intercept )
            gdImageLine( im, 
                         y_intercept,        y0,
                         y_intercept+xdepth, y0-ydepth, clrshd ); /* edging */
        gdImageLine( im, x1, y0, x1+xdepth, y0-ydepth, clrshd );  /* edging */
        gdImageLine( im, x2, y0, x2+xdepth, y0-ydepth, clrshd );  /* edging */

        /* SET_3D_POLY( poly, x2, x2, y0, y2, xdepth, ydepth ); */ /*  side */
        /* gdImageFilledPolygon( im, poly, 4, clrshd ); */

        gdImageLine( im, x1, y1, x1+xdepth, y1-ydepth, clrshd );  /* edging */
        gdImageLine( im, x2, y2, x2+xdepth, y2-ydepth, clrshd );  /* edging */
    }

    if( y1 == y2 )                                              /* bar rect */
        SET_RECT( poly, x1, x2, y0, y1 );                       /* front */
    else
    {
        poly[0].x = x1; poly[0].y = y0;
        poly[1].x = x2; poly[1].y = y0;
        poly[2].x = x2; poly[2].y = y2;
        poly[3].x = x1; poly[3].y = y1;
    }
    gdImageFilledPolygon( im, poly, 4, clr );

    gdImageLine( im, x1, y0, x2, y0, clrshd );  /* edging along y0 */

    if( (xdepth || ydepth) &&                   /* front edging only on 3D */
        (y1<y0 || y2<y0) )                      /* and only above y0       */
    {
        if( y1 > y0 && y2 < y0 )                /* line crosses from below y0*/
            gdImageLine( im, y_intercept, y0, x2, y2, clrshd );
        else
          if( y1 < y0 && y2 > y0 )              /* line crosses from above y0*/
            gdImageLine( im, x1, y1, y_intercept, y0, clrshd );
          else                                  /* completely above */
            gdImageLine( im, x1, y1, x2, y2, clrshd );
    }
}

/**********************************************************************
 *                       draw_3d_bar()
 **********************************************************************/
/* ------------------------------------------------------------------------- */
/* vals in pixels */
/* ref is front plane */
static void
draw_3d_bar( gdImagePtr         im,
             int                x1,

             int                x2,
             int                y0,
             int                yhigh,
             int                xdepth,
             int                ydepth,
             int                clr,
             int                clrshd,
             char               flip_axis) /* FALSE= vertical bar,
                                              TRUE = horizontal bar */
{
#define SET_3D_BAR( gdp, x1, x2, y1, y2, xoff, yoff )                       \
                                gdp[0].x  = x1,        gdp[0].y = y1,       \
                                gdp[1].x  = x1+(xoff), gdp[1].y = y1-yoff,  \
                                gdp[2].x  = x2+(xoff), gdp[2].y = y2-yoff,  \
                                gdp[3].x  = x2,        gdp[3].y = y2

    gdPoint     poly[4];

    if ( !flip_axis )
    {   
        int     ycap = MIN( y0, yhigh );         /* up-side-down bars */

        if ( xdepth || ydepth )
        {
            if( y0 != yhigh )                   /* 0 height? */
            {
                SET_3D_BAR( poly, x2, x2, y0, yhigh, xdepth, ydepth ); /* side */
                gdImageFilledPolygon( im, poly, 4, clrshd );
            }

            SET_3D_BAR( poly, x1, x2, ycap, ycap, xdepth, ydepth );      /* top */
            gdImageFilledPolygon( im, poly, 4, clr );
        }
    
        SET_RECT( poly, x1, x2, y0, yhigh );                           /* front */
        gdImageFilledPolygon( im, poly, 4, clr );

        if( xdepth || ydepth )
            gdImageLine( im, x1, ycap, x2, ycap, clrshd );
    }
    else
    {
        int     xcap = MAX( x1, x2 );         /* left-to-right bars */
        if ( xdepth || ydepth )
        {
            if( x1 != x2 )                   /* 0 length? */
            {
                SET_3D_BAR( poly, x1, x2, yhigh, yhigh, xdepth, ydepth ); /* top side */
                gdImageFilledPolygon( im, poly, 4, clrshd );
            }

            SET_3D_BAR( poly, xcap, xcap, y0, yhigh, xdepth, ydepth );      /* cap (right end) */
            gdImageFilledPolygon( im, poly, 4, clr );
        }
    
        SET_RECT( poly, x1, x2, y0, yhigh );                           /* front */
        gdImageFilledPolygon( im, poly, 4, clr );

        if( xdepth || ydepth )
            gdImageLine( im, xcap, y0, xcap, yhigh, clrshd );

    }

}


/**********************************************************************
 *                       GDCImageLine()
 *
 * Draw a 2D line, 2 pixels wide, making sure GD line styles will 
 * work properly.
 **********************************************************************/
static void GDCImageLine(gdImagePtr im, int x1, int y1, int x2, int y2, int color)
{
    int nStylePos, xOffset=0, yOffset=0;

    if ((x2-x1) > (y2-y1))
        yOffset=1;
    else
        xOffset=1;

    /* Save current GD style offset and restore it before drawing the
     * second line segment
     */
    nStylePos = im->stylePos;
    gdImageLine(im, x1, y1, x2, y2, color);
    im->stylePos = nStylePos;
    gdImageLine(im, x1+xOffset, y1+yOffset, x2+xOffset, y2+yOffset, color);
}


/* ------------------------------------------------------------------------- */
struct BS { float y1; float y2; int clr; int shclr; };
static int barcmpr( const void *a, const void *b )
	{ if( ((struct BS*)a)->y2 < ((struct BS*)b)->y2 ) return -1;
	  if( ((struct BS*)a)->y2 > ((struct BS*)b)->y2 ) return 1;
	  return 0; }

/* ------------------------------------------------------------------------- */
// simple two-point linear interpolation
// attempts between first, then nearest
void
do_interpolations( int	 num_points,
                   int	 interp_point,
                   float vals[] )
{
    int     i, j;
    float   v1 = GDC_NOVALUE,
        v2 = GDC_NOVALUE;
    int         p1 = -1,
        p2 = -1;

    // find backwards
    for( i=interp_point-1; i>=0 && p1==-1; --i )
        if( vals[i] != GDC_NOVALUE && vals[i] != GDC_INTERP_VALUE )
        {
            v1 = vals[i];
            p1 = i;
        }
    // find forwards
    for( j=interp_point+1; j<num_points && p2==-1; ++j )
        if( vals[j] != GDC_NOVALUE && vals[j] != GDC_INTERP_VALUE )
        {
            v2 = vals[j];
            p2 = j;
        }
    // no forward sample, find backwards
    for( ; i>=0 && p2==-1; --i )
        if( vals[i] != GDC_NOVALUE && vals[i] != GDC_INTERP_VALUE )
        {
            v2 = vals[i];
            p2 = i;
        }
    // no backwards sample, find forwards
    for( ; j<num_points && p1==-1; ++j )
        if( vals[j] != GDC_NOVALUE && vals[j] != GDC_INTERP_VALUE )
        {
            v1 = vals[j];
            p1 = j;
        }
    if( p1==-1 || p2==-1 ||     // need both
        p1 == p2 )              // idiot
    {
        vals[interp_point] = GDC_NOVALUE;
        return;
    }

    // Point-slope formula
    vals[interp_point] = ((v2-v1)/(float)(p2-p1)) * (float)(interp_point-p1) + v1;
    return;
}

/**********************************************************************
 *                       _GDCXYChartGIF()
 *
 * This is the function where all the magic(!) happens.
 *
 * little/no error checking  0:    ok,
 *                           -ret: error no graph output
 *                           ret:  error graph out
 *
 * No check for a gif that's too small to fit needed info (labels, etc), 
 * could core dump.
 * watch out for # params and array sizes==num_points
 **********************************************************************/

int  _GDCXYChartGIF( GDCChartPtr psGDC,
                     FILE        *img_fptr)
{
    int         i, j;

    int         graphwidth;
    int         grapheight;
    gdImagePtr  im;
    gdImagePtr  bg_img = NULL;

    DEF_CLR_ALLOCATE_VARS;

    float       xorig, yorig, vyorig, false_xorig;
    float       yscl     = 0.0;
    float       vyscl    = 0.0;
    float       xscl     = 0.0;
    float       vhighest = -MAXFLOAT;
    float       vlowest  = MAXFLOAT;
    float       highest  = -MAXFLOAT;
    float       lowest   = MAXFLOAT;
    float       val_highest = -MAXFLOAT;

    char        do_vol = ( psGDC->type == GDC_COMBO_HLC_BAR || /* aka: combo */
                           psGDC->type == GDC_COMBO_HLC_AREA  ||
                           psGDC->type == GDC_COMBO_HLC_LINE  ||
                           psGDC->type == GDC_COMBO_LINE_BAR  ||
                           psGDC->type == GDC_COMBO_LINE_AREA ||
                           psGDC->type == GDC_COMBO_LINE_LINE ||
                           psGDC->type == GDC_COMBO_LINE_HLC  ||
                           psGDC->type == GDC_3DCOMBO_HLC_BAR ||
                           psGDC->type == GDC_3DCOMBO_HLC_AREA||
                           psGDC->type == GDC_3DCOMBO_HLC_LINE||
                           psGDC->type == GDC_3DCOMBO_LINE_HLC||
                           psGDC->type == GDC_3DCOMBO_LINE_BAR||
                           psGDC->type == GDC_3DCOMBO_LINE_AREA||
                           psGDC->type == GDC_3DCOMBO_LINE_LINE );
    char        threeD = ( psGDC->type == GDC_3DAREA          ||
                           psGDC->type == GDC_3DLINE          ||
                           psGDC->type == GDC_3DBAR           ||
                           psGDC->type == GDC_3DFLOATINGBAR   ||
                           psGDC->type == GDC_3DHBAR          ||
                           psGDC->type == GDC_3DH2BAR         ||
                           psGDC->type == GDC_3DHILOCLOSE     ||
                           psGDC->type == GDC_3DCOMBO_HLC_BAR ||
                           psGDC->type == GDC_3DCOMBO_HLC_AREA||
                           psGDC->type == GDC_3DCOMBO_HLC_LINE||
                           psGDC->type == GDC_3DCOMBO_LINE_BAR||
                           psGDC->type == GDC_3DCOMBO_LINE_AREA||
                           psGDC->type == GDC_3DCOMBO_LINE_HLC||
                           psGDC->type == GDC_3DCOMBO_LINE_LINE );
    char        do_bar = ( psGDC->type == GDC_3DBAR  ||  /* offset X objects to leave*/
                           psGDC->type == GDC_BAR    ||  /*  room at X(0) and X(n)   */
                           psGDC->type == GDC_3DFLOATINGBAR ||
                           psGDC->type == GDC_FLOATINGBAR ||
                           psGDC->type == GDC_3DHBAR ||
                           psGDC->type == GDC_HBAR   ||
                           psGDC->type == GDC_3DH2BAR||
                           psGDC->type == GDC_H2BAR   );
    char        do_h2bar = ( psGDC->type == GDC_3DH2BAR ||   /* Bars to the left and to */
                             psGDC->type == GDC_H2BAR );     /* the right */
    char        flip_axis = ( psGDC->type == GDC_3DHBAR ||   /* Y axis horizontal   */
                              psGDC->type == GDC_HBAR   ||   /* and X axis vertical */
                              psGDC->type == GDC_3DH2BAR||
                              psGDC->type == GDC_H2BAR );
    char        do_ylbl_fractions =             /* %f format not given, or */
                        ( !psGDC->ylabel_fmt ||   /*  format doesn't have a %,g,e,E,f or F */
                           strlen(psGDC->ylabel_fmt) == strcspn(psGDC->ylabel_fmt,"%geEfF") );
    float       ylbl_interval  = 0.0;
    int         num_lf_xlbls   = 0;
    int         xdepth_3Dtotal = 0;
    int         ydepth_3Dtotal = 0;
    int         xdepth_3D      = 0;     /* affects PX() */
    int         ydepth_3D      = 0;     /* affects PY() and PV() */
    int         hlf_barwdth    = 0;     /* half bar widths */
    int         hlf_hlccapwdth = 0;     /* half cap widths for HLC_I_CAP and DIAMOND */
    int         annote_len     = 0,
                annote_hgt     = 0;

    char        set_depth;

    /* args */
    int         setno = 0;              /* affects PX() and PY() */

    float       **uvals;
    float       *uvol;  
    char        **xlbl;

    float       *log_ylbl;      /* Y labels for log scale charts */
    int         log_num_ylbl;
    float       bars_ybase = 0.0;

    int         BGColor,
                LineColor,
                PlotColor,
                GridColor,
                VolColor,
                /*ExtVolColor[num_points], */
                /*ExtColor[num_sets][num_points],*/
                ThumbDColor,
                ThumbLblColor,
                ThumbUColor,
/*              ArrowDColor,*/
/*              ArrowUColor,*/
                AnnoteColor;
    DEF_ARRAY(int, ExtVolColor, psGDC->num_points);
    DEF_2DARRAY(int, ExtColor, psGDC->num_sets, psGDC->num_points);
                                                                    /* shade colors only with 3D */
/*  int         ExtColorShd[threeD?1:num_sets][threeD?1:num_points]; // compiler limitation */
/* int           ExtColorShd[num_sets][num_points]; */
    DEF_2DARRAY(int, ExtColorShd, psGDC->num_sets, psGDC->num_points);

    DEF_ARRAY(GDCLineStyle, LineStyles, psGDC->num_sets);


    ALLOC_ARRAY(int, ExtVolColor, psGDC->num_points);
    ALLOC_2DARRAY(int, ExtColor, psGDC->num_sets, psGDC->num_points);
    ALLOC_2DARRAY(int, ExtColorShd, psGDC->num_sets, psGDC->num_points);
    LineStyles = NULL;

    set_depth = ( psGDC->stack_type == GDC_STACK_DEPTH )? psGDC->num_groups:1;

    /* To simplify the source code, use temp vars for 
     * uvals[][], uvol[], and xlbl */
    uvals = psGDC->uvals;
    uvol = psGDC->uvol;
    xlbl = psGDC->xlbl;

    log_ylbl = NULL;
    log_num_ylbl = 0;

    /* idiot checks */
    if( psGDC->imgwidth<=0 || psGDC->imgheight<=0 || (!img_fptr && psGDC->generate_img) )
    {
        FREE_ARRAY(ExtVolColor);
        FREE_ARRAY(ExtColor);
        FREE_ARRAY(ExtColorShd);
        return -1;
    }
    if( psGDC->num_points <= 0 )
        {
        GDC_out_err( psGDC, psGDC->imgwidth, psGDC->imgheight, img_fptr, 
                     psGDC->BGColor, psGDC->LineColor, "No Data Available" );
        FREE_ARRAY(ExtVolColor);
        FREE_ARRAY(ExtColor);
        FREE_ARRAY(ExtColorShd);
        return 1;
        }

    GDC_load_font_conversions();
    if( psGDC->thumbnail )
        {
        psGDC->grid = FALSE;
        psGDC->xaxis = FALSE;
        psGDC->yaxis = FALSE;
        }

    /* ----- calculate interpretations first ----- */
    if( psGDC->interpolations )
    {
        for( i=0; i<psGDC->num_sets; ++i )
            for( j=0; j<psGDC->num_points; ++j )
                if( uvals[i][j] == GDC_INTERP_VALUE )
                {
                    do_interpolations( psGDC->num_points, j, uvals[i] );
                }
        if( do_vol )
            for( j=0; j<psGDC->num_points; ++j )
                if( uvol[j] == GDC_INTERP_VALUE )
                {
                    do_interpolations( psGDC->num_points, j, uvol );
                }
    }

    /* ----- highest & lowest values ----- */
    if( psGDC->stack_type == GDC_STACK_SUM )       /* need to walk sideways */
        for( j=0; j<psGDC->num_points; ++j )
            {
            float   set_sum = 0.0;
            for( i=0; i< psGDC->num_sets; ++i )
                if( uvals[i][j] != GDC_NOVALUE )
                    {
                    set_sum += uvals[i][j];
                    highest = MAX( highest, set_sum );
                    lowest  = MIN( lowest,  set_sum );
                    }
            }
    else
    if( psGDC->stack_type == GDC_STACK_LAYER )     /* need to walk sideways */
        for( j=0; j<psGDC->num_points; ++j )
            {
            float   neg_set_sum = 0.0,
                    pos_set_sum = 0.0;
            for( i=0; i< psGDC->num_sets; ++i )
                if( uvals[i][j] != GDC_NOVALUE )
                    if( uvals[i][j] < 0.0 )
                        neg_set_sum += uvals[i][j];
                    else
                        pos_set_sum += uvals[i][j];
            lowest  = MIN( lowest,  MIN(neg_set_sum,pos_set_sum) );
            highest = MAX( highest, MAX(neg_set_sum,pos_set_sum) );
            }
    else
        for( i=0; i< psGDC->num_sets; ++i )
            for( j=0; j<psGDC->num_points; ++j )
                if( uvals[i][j] != GDC_NOVALUE )
                    {
                    highest = MAX( uvals[i][j], highest );
                    lowest  = MIN( uvals[i][j], lowest );
                    }
    if( psGDC->scatter )
      for( i=0; i<psGDC->num_scatter_pts; ++i )
        {
        highest = MAX( (psGDC->scatter+i)->val, highest );
        lowest  = MIN( (psGDC->scatter+i)->val, lowest  );
        }
    if( do_vol )                                /* for now only one combo set allowed */
        {
        /* vhighest = 1.0;*/
        /* vlowest  = 0.0;*/
        if( psGDC->type == GDC_COMBO_LINE_HLC    ||
            psGDC->type == GDC_3DCOMBO_LINE_HLC )
            {
            for( j=0; j<(psGDC->num_points*3); ++j )
                if( uvol[j] != GDC_NOVALUE )
                    {
                    vhighest = MAX( uvol[j], vhighest );
                    vlowest  = MIN( uvol[j], vlowest );
                    }
            }
        else
            {
            for( j=0; j<psGDC->num_points; ++j )
                if( uvol[j] != GDC_NOVALUE )
                    {
                    vhighest = MAX( uvol[j], vhighest );
                    vlowest  = MIN( uvol[j], vlowest );
                    }
            }

        if( vhighest == -MAXFLOAT )         /* no values */
            vhighest = 1.0;                     /* for scaling, need a range */
        if( vlowest == MAXFLOAT )
            vlowest = 0.0;
        if( psGDC->type == GDC_COMBO_LINE_BAR    ||
            psGDC->type == GDC_COMBO_HLC_BAR     ||
            psGDC->type == GDC_COMBO_LINE_AREA   ||
            psGDC->type == GDC_COMBO_HLC_AREA    ||
            psGDC->type == GDC_COMBO_HLC_LINE    ||
            psGDC->type == GDC_3DCOMBO_LINE_BAR  ||
            psGDC->type == GDC_3DCOMBO_LINE_AREA ||
            psGDC->type == GDC_3DCOMBO_HLC_BAR   ||
            psGDC->type == GDC_3DCOMBO_HLC_LINE  ||
            psGDC->type == GDC_3DCOMBO_HLC_AREA )
            if( vhighest < 0.0 )
                vhighest = 0.0;
            else
                if( vlowest > 0.0  )
                    vlowest = 0.0;     /* bar, area should always start at 0 */

        if( !psGDC->yaxis2 )
            {
                /* yaxis2 not requested: use same min/max for both sets */
                lowest = vlowest = MIN(lowest, vlowest);
                highest = vhighest = MAX(highest, vhighest);
            }
        }

    if( lowest == MAXFLOAT )
        lowest = (psGDC->yscale_type==GDC_SCALE_LOG)? 1.0:0.0;
    if( highest == -MAXFLOAT )
        highest = (psGDC->yscale_type==GDC_SCALE_LOG)?10.0:1.0;     /* need a range */

    if( psGDC->requested_ymin != GDC_NOVALUE && psGDC->requested_ymin < lowest )
        lowest = psGDC->requested_ymin;
    if( psGDC->requested_ymax != GDC_NOVALUE && psGDC->requested_ymax > highest )
    {
        val_highest = highest;
        highest = psGDC->requested_ymax;
    }
    
    if( psGDC->requested_ymin2 != GDC_NOVALUE && psGDC->requested_ymin2 < vlowest )
        vlowest = psGDC->requested_ymin2;
    if( psGDC->requested_ymax2 != GDC_NOVALUE && psGDC->requested_ymax2 > vhighest )
        vhighest = psGDC->requested_ymax2;

    if( psGDC->type == GDC_AREA   ||           /* bars and area should always start at 0 */
        psGDC->type == GDC_BAR    ||           /* or at 1 for log scale */
        psGDC->type == GDC_3DBAR  ||
        psGDC->type == GDC_3DAREA ||
        psGDC->type == GDC_HBAR   ||
        psGDC->type == GDC_3DHBAR ||
        psGDC->type == GDC_H2BAR  ||
        psGDC->type == GDC_3DH2BAR)
    {
        bars_ybase = 0.0;  /* Set a default value (0.0 for logs is 1 in reality) */

        if( highest < 0.0 )
        {
            highest = (psGDC->yscale_type==GDC_SCALE_LOG)? MAX(-1.0, pow(10, ceil(log10(highest)))) : 0.0;
            bars_ybase = (psGDC->yscale_type==GDC_SCALE_LOG)? log10(highest) : highest;
        }
        else
        if( lowest > 0.0 )                     /* negs should be drawn from 0*/
        {
            lowest = (psGDC->yscale_type==GDC_SCALE_LOG)? MIN(1.0, pow(10, floor(log10(lowest)))) : 0.0;
            bars_ybase = (psGDC->yscale_type==GDC_SCALE_LOG)? log10(lowest) : lowest;
        }
    }


    /* ----- graph height and width within the img height width ----- */
    /* grapheight/height is the actual size of the scalable graph, 
     * including the axis labels.
     * When flip_axis = TRUE:
     *   graphwidth becomes the vertical size of the chart
     *   grapheight becomes the horizontal size of the chart.
     */
    {
    int title_hgt  = psGDC->title? 2               /* title? horizontal text line(s) */
                                + GDC_cnt_nl(psGDC->title,(int*)NULL)*GDC_fontc[psGDC->title_size].h
                                + 2:
                                2;
    int ltitle_hgt = (do_h2bar && psGDC->ltitle)? (2 + GDC_cnt_nl(psGDC->ltitle,(int*)NULL)*GDC_fontc[psGDC->title_size].h) : 0;
    int rtitle_hgt = (do_h2bar && psGDC->rtitle)? (2 + GDC_cnt_nl(psGDC->rtitle,(int*)NULL)*GDC_fontc[psGDC->title_size].h) : 0;
    int xlabel_hgt = 0;
    int xtitle_hgt = psGDC->xtitle? 1+GDC_fontc[psGDC->xtitle_size].h+1: 0;
    int ytitle_hgt = psGDC->ytitle? 1+GDC_fontc[psGDC->ytitle_size].h+1: 0;
    int vtitle_hgt = do_vol&&psGDC->ytitle2? 1+GDC_fontc[psGDC->ytitle_size].h+1: 0;
    int ylabel_wth = 0;
    int vlabel_wth = 0;

    int xtics       = psGDC->ticks && (psGDC->grid||psGDC->xaxis)? 1+2: 0;
    int ytics       = psGDC->ticks && (psGDC->grid||psGDC->yaxis)? 1+3: 0;
    int vtics       = psGDC->ticks && (psGDC->yaxis&&do_vol)? 3+1: 0;


#define HYP_DEPTH   ( (double)((psGDC->imgwidth+psGDC->imgheight)/2) * ((double)psGDC->_3d_depth)/100.0 )
#define RAD_DEPTH   ( (double)psGDC->_3d_angle*2*M_PI/360 )
    xdepth_3D      = threeD? (int)( (flip_axis? sin(RAD_DEPTH): cos(RAD_DEPTH)) * HYP_DEPTH ): 0;
    ydepth_3D      = threeD? (int)( (flip_axis? cos(RAD_DEPTH): sin(RAD_DEPTH)) * HYP_DEPTH ): 0;
    xdepth_3Dtotal = xdepth_3D*set_depth;
    ydepth_3Dtotal = ydepth_3D*set_depth;
    annote_hgt = psGDC->annotation && *(psGDC->annotation->note)?
                    1 +                                         /* space to note */
                    (1+GDC_fontc[psGDC->annotation_font].h) *      /* number of '\n' substrs */
                    GDC_cnt_nl(psGDC->annotation->note,&annote_len) +
                    1 +                                         /* space under note */
                    2: 0;                                       /* space to chart */
    annote_len *= GDC_fontc[psGDC->annotation_font].w;

    if( psGDC->xaxis && xlbl )
        {
        int biggest     = -MAXINT;

        for( i=0; i<psGDC->num_points; ++i )
            {
            int     len = 0;
                                                               /* longest "...\n" segment */
            for( len=0, j=0; xlbl[i][j]; ++len, ++j )
                if( xlbl[i][j] == '\n' )
                    {
                    biggest = MAX( len, biggest );
                    ++num_lf_xlbls;
                    len = 0;
                    }
            biggest = MAX( len, biggest );                     /* last seg */
            }
        xlabel_hgt = 1+ biggest*GDC_fontc[psGDC->xaxisfont_size].w +1;
        }
    if ( !flip_axis )
        grapheight = psGDC->imgheight - ( xtics          +
                                          xtitle_hgt     +
                                          xlabel_hgt     +
                                          title_hgt      +
                                          annote_hgt     +
                                          ydepth_3Dtotal +
                                          2 );
    else
        grapheight = psGDC->imgwidth - ( xtics          +
                                         xtitle_hgt     +
                                         xlabel_hgt     +
                                         xdepth_3Dtotal +
                                         2 );
    if( psGDC->hard_size && psGDC->hard_grapheight )              /* user wants to use his */
        grapheight = psGDC->hard_grapheight;
    psGDC->hard_grapheight = grapheight;

    /* ----- Can add Val label at the top of the chart ----- */
    if ( psGDC->bar_labels && (psGDC->requested_ymax == GDC_NOVALUE) 
         && do_bar && !do_h2bar)
    {
        float   label_value, label_width, scale, high;
        char    foo[32];

        if( !do_bar || do_h2bar)
        {
            GDC_out_err( psGDC, psGDC->imgwidth, psGDC->imgheight,
                         img_fptr,
                         psGDC->BGColor, psGDC->LineColor,
                         "BarLabels not implemented for this chart type." );
            FREE_ARRAY(ExtVolColor);
            FREE_ARRAY(ExtColor);
            FREE_ARRAY(ExtColorShd);

            return 2;
        }

        high = (val_highest == -MAXFLOAT? highest : val_highest);

        /* Get the label value */
        sprintf( foo, psGDC->bar_label_fmt? psGDC->bar_label_fmt : " %.0f ", high );
        label_width = strlen( foo ) * GDC_fontc[psGDC->xaxisfont_size].w;

        scale = ((float)grapheight - label_width) / (float)(highest-lowest);

        label_value = ( label_width / scale);

        if ( ( high + label_value ) > highest )
            highest = high + label_value;
    }

                                                            /* before width can be known... */
    /* ----- y labels intervals ----- */
    if (psGDC->yscale_type == GDC_SCALE_LINEAR)
    {
        float   tmp_highest;
                                                            /* possible y gridline points */
        float   ypoints[] = { 1.0/64.0, 1.0/32.0, 1.0/16.0, 1.0/8.0, 1.0/4.0, 1.0/2.0,
                              1.0,      2.0,      3.0,      5.0,     10.0,    25.0,
                              50.0,     100.0,    250.0,    500.0,   1000.0,  2500,    5000.0,
                              10000.0,  25000.0,  50000.0,  100000.0,500000.0,1000000, 5000000,
                              10000000 };
        #define NUM_YPOINTS (sizeof(ypoints) / sizeof(float))
        int     max_num_ylbls;
        int     longest_ylblen = 0;
                                                            /* maximum y lables that'll fit... */
        max_num_ylbls = grapheight / (3+GDC_fontc[psGDC->yaxisfont_size==GDC_TINY? psGDC->yaxisfont_size+1:
                                                                            psGDC->yaxisfont_size].h);
        if( max_num_ylbls < 3 )
        {
            /* gdImageDestroy(im);      haven't yet created it */
            GDC_out_err( psGDC, psGDC->imgwidth, psGDC->imgheight,
                         img_fptr,
                         psGDC->BGColor, psGDC->LineColor,
                         "Insificient Height" );
            FREE_ARRAY(ExtVolColor);
            FREE_ARRAY(ExtColor);
            FREE_ARRAY(ExtColorShd);

            return 2;
        }

                                                        /* one "space" interval above + below */
        for( i=1; i<NUM_YPOINTS; ++i )
            /* if( ypoints[i] > ylbl_interval ) */
            /*  break; */
            if( (highest-lowest)*(do_h2bar?2:1)/ypoints[i] < ((float)max_num_ylbls-(1.0+1.0))
                                                             * (float)psGDC->ylabel_density/100.0 )
                break;
        /* gotta go through the above loop to catch the 'tweeners :-| */

        ylbl_interval = psGDC->requested_yinterval != GDC_NOVALUE &&
                        psGDC->requested_yinterval > ypoints[i-1]?   psGDC->requested_yinterval:
                                                              ypoints[i-1];

                                                        /* perform floating point remainders */
                                                        /* gonculate largest interval-point < lowest */
        if( lowest != 0.0 &&
            lowest != psGDC->requested_ymin )
        {
            if( lowest < 0.0 )
                lowest -= ylbl_interval;
            /* lowest = (lowest-ypoints[0]) -
             *           ( ( ((lowest-ypoints[0])/ylbl_interval)*ylbl_interval ) -
             *              ( (float)((int)((lowest-ypoints[0])/ylbl_interval))*ylbl_interval ) );
             */
            lowest = ylbl_interval * (float)(int)((lowest-ypoints[0])/ylbl_interval);
        }
                                                        /* find smallest interval-point > highest */
        tmp_highest = lowest;
        do  /* while( (tmp_highest += ylbl_interval) <= highest ) */
        {
            int     nmrtr, dmntr, whole;
            int     lbl_len;
            char    foo[32];
            if( psGDC->yaxis )
            {                                           /* XPG2 compatibility */
                sprintf( foo, do_ylbl_fractions? "%.0f": psGDC->ylabel_fmt, tmp_highest );
                lbl_len = ylbl_interval<1.0? strlen( GDC_price_to_str(tmp_highest,
                                                                      &nmrtr,
                                                                      &dmntr,
                                                                      &whole,
                                                                      do_ylbl_fractions? NULL: psGDC->ylabel_fmt) ):
                                              strlen( foo );
                longest_ylblen = MAX( longest_ylblen, lbl_len );
            }
        } while( (tmp_highest += ylbl_interval) <= highest );
        ylabel_wth = longest_ylblen * GDC_fontc[psGDC->yaxisfont_size].w;
        highest = psGDC->requested_ymax==GDC_NOVALUE? tmp_highest:
                                               MAX( psGDC->requested_ymax, highest );

        if( do_vol )         /* __TODO__ we should try to get nice labels and a nice interval */
        {
            float   num_yintrvls = (highest-lowest) / ylbl_interval;
                                                            /* no skyscrapers */
            if( vhighest != 0.0 && psGDC->requested_ymax2 != vhighest)
                vhighest += (vhighest-vlowest) / (num_yintrvls*2.0);
            if( vlowest != 0.0 && psGDC->requested_ymin2 != vlowest)
                vlowest -= (vhighest-vlowest) / (num_yintrvls*2.0);

            if( psGDC->yaxis2 )
            {
                char    svlongest[32];
                int     lbl_len_low  = sprintf( svlongest, psGDC->ylabel2_fmt? psGDC->ylabel2_fmt: "%.0f", vlowest );
                int     lbl_len_high = sprintf( svlongest, psGDC->ylabel2_fmt? psGDC->ylabel2_fmt: "%.0f", vhighest );
                vlabel_wth = 1
                         + MAX( lbl_len_low,lbl_len_high ) * GDC_fontc[psGDC->yaxisfont_size].w;
            }
        }

    }
    else if (psGDC->yscale_type == GDC_SCALE_LOG)
    {
        /* LOG Scale... the y labels will be based on powers of 10.
         * Values <= 0 cannot be plotted.
         */
        float   log_highest, log_lowest;
        int     max_num_ylbls;
        int     longest_ylblen = 0;

        if (do_vol)
        {
            GDC_out_err( psGDC, psGDC->imgwidth, psGDC->imgheight,
                         img_fptr,
                         psGDC->BGColor, psGDC->LineColor,
                         "Log scale with COMBO charts not supported yet." );
            return 2;
        }

        if (lowest <= 0.0 || highest <= 0)
        {
            GDC_out_err( psGDC, psGDC->imgwidth, psGDC->imgheight,
                         img_fptr,
                         psGDC->BGColor, psGDC->LineColor,
                         "Log scale chart cannot have values <= 0." );

            FREE_ARRAY(ExtVolColor);
            FREE_ARRAY(ExtColor);
            FREE_ARRAY(ExtColorShd);

            return 2;
        }

        log_highest = (int)(ceil((float)log10(highest)));
        log_lowest = (int)(floor((float)log10(lowest)));
        ylbl_interval = 1.0; 
        log_num_ylbl = log_highest - log_lowest;

        /* Build an array of nice label values.  
         * This could certainly be improved, but for now, it will be
         * powers of 10, and either 1 or 2 sub-divisions between them
         * e.g. {10, 20, 50, 100} or 10, 30, 100}
         */
        /* maximum y labels that'll fit... */
        max_num_ylbls = grapheight / (3+GDC_fontc[psGDC->yaxisfont_size==GDC_TINY? psGDC->yaxisfont_size+1:
                                                                            psGDC->yaxisfont_size].h);
        max_num_ylbls = max_num_ylbls * (psGDC->ylabel_density/100.0);

        if (log_num_ylbl*3 <= max_num_ylbls)
        {
            /* 2 sub-divisions between each power of 10:  10, 20, 50, 100 */
            log_num_ylbl *= 3;
            log_ylbl = (float*)malloc(log_num_ylbl*sizeof(float));
            for(i=0; i < log_num_ylbl; )
            {
                float fval;
                fval = pow(10, log_lowest+i/3 );
                log_ylbl[i++] = fval;
                log_ylbl[i++] = fval*2.5;
                log_ylbl[i++] = fval*5.0;
            }

        }
        else if (log_num_ylbl*2 <= max_num_ylbls)
        {
            /* 1 sub-division between each power of 10:  10, 30, 100 */
            log_num_ylbl *= 2;
            log_ylbl = (float*)malloc(log_num_ylbl*sizeof(float));
            for(i=0; i < log_num_ylbl; )
            {
                float fval;
                fval = pow(10, log_lowest+i/2 );
                log_ylbl[i++] = fval;
                log_ylbl[i++] = fval*3.0;
            }
        }
        else
        {
            /* We already have too many...
             * __TODO__ remove some powers of 10 */

            log_ylbl = (float*)malloc(log_num_ylbl*sizeof(float));
            for(i=0; i < log_num_ylbl; )
            {
                float fval;
                fval = pow(10, log_lowest+i );
                log_ylbl[i++] = fval;
            }
        }

        /* Establish the longest label's size */
        longest_ylblen = MAX( ABS(log_lowest), ABS(log_highest) ) + 1;
        if (log_lowest < 0.0)
            longest_ylblen += 2;  /* for the "0." before the decimal values */
        ylabel_wth = longest_ylblen * GDC_fontc[psGDC->yaxisfont_size].w;

        do_ylbl_fractions = FALSE;
        if (psGDC->ylabel_fmt == NULL)
        psGDC->ylabel_fmt = "%g";

        /* OK, here's the trick... instead of adding log10() calls
         * everywhere in the rest of the function, we'll convert
         * all values to their log10() value, and continue just
         * as if we were creating a linear scale chart.
         *
         * The side effect of this is that the arrays of values
         * passed by the user will be modified... 
         * This could also lead to problems with static arrays
         * with some compilers (i.e. Seg. Fault)
         */
        highest = log_highest;
        lowest = log_lowest;

        for( j=psGDC->num_sets -1 ; j >= 0;  j-- )
        {
            for( i=0; i<psGDC->num_points; ++i )
                if( uvals[j][i] != GDC_NOVALUE )
                    uvals[j][i] = log10(uvals[j][i]);
        }

    }
    else
    {
        GDC_out_err( psGDC, psGDC->imgwidth, psGDC->imgheight,
                     img_fptr,
                     psGDC->BGColor, psGDC->LineColor,
                     "Unsupported scale type." );
            FREE_ARRAY(ExtVolColor);
            FREE_ARRAY(ExtColor);
            FREE_ARRAY(ExtColorShd);

            return 2;
    }


    /* graphwidth: width of the chart, including axis annotations.
     * This value becomes the vertical size of the chart when flip_axis = TRUE
     */
    if ( !flip_axis )
        graphwidth = psGDC->imgwidth - ( (psGDC->hard_size && psGDC->hard_xorig)? psGDC->hard_xorig:
                                                                   ( ytitle_hgt +
                                                                     ylabel_wth +
                                                                     ytics + 2) );
    else
        graphwidth = psGDC->imgheight - ( (psGDC->hard_size && psGDC->hard_yorig)? psGDC->hard_yorig:
                                                                   ( title_hgt +
                                                                     MAX(ltitle_hgt, rtitle_hgt) +
                                                                     annote_hgt +
                                                                     ytitle_hgt +
                                                                     ylabel_wth +
                                                                     ytics + 2) );
    graphwidth -= ( vtics + vtitle_hgt + vlabel_wth + xdepth_3Dtotal );

    if( psGDC->hard_size && psGDC->hard_graphwidth )              /* user wants to use his */
        graphwidth = psGDC->hard_graphwidth;
    psGDC->hard_graphwidth = graphwidth;

    /* ----- scale to img size ----- */
    /* offset to 0 at lower left (where it should be) */
    if ( !flip_axis )
    {
/*        xscl = (float)(graphwidth-xdepth_3Dtotal) / (float)(psGDC->num_points + (do_bar?2:0));*/
        xscl = (float)(graphwidth-xdepth_3Dtotal) / (float)(psGDC->num_points + (do_bar?1:0));
        yscl = -((float)grapheight) / (float)(highest-lowest);
    }
    else
    {
/*        yscl = -(float)(graphwidth-ydepth_3Dtotal) / (float)(psGDC->num_points + (do_bar?2:0)); */
        yscl = -(float)(graphwidth-ydepth_3Dtotal) / (float)(psGDC->num_points + (do_bar?1:0));
        xscl = ((float)grapheight) / (float)((highest-lowest)*(do_h2bar?2:1));
    }
    if( do_vol )
        {
        float   hilow_diff = vhighest-vlowest==0.0? 1.0: vhighest-vlowest;

        vyscl = -((float)grapheight) / hilow_diff;
        vyorig = (float)grapheight
                 + ABS(vyscl) * MIN(vlowest,vhighest)
                 + ydepth_3Dtotal
                 + title_hgt
                 + annote_hgt;
        }

    if ( !flip_axis )
        {
        xorig = (float)( psGDC->imgwidth - ( graphwidth +
                                             vtitle_hgt +
                                             vtics      +
                                             vlabel_wth + 2) );
        if( psGDC->hard_size && psGDC->hard_xorig )
            xorig = psGDC->hard_xorig;
        false_xorig = xorig;
/*  yorig = (float)grapheight + ABS(yscl * lowest) + ydepth_3Dtotal + title_hgt; */
        yorig = (float)grapheight
                       + ABS(yscl) * MIN(lowest,highest)
                       + ydepth_3Dtotal
                       + title_hgt
                       + annote_hgt;
/*????  if( psGDC->hard_size && psGDC->hard_yorig ) */                  /* vyorig too? */
/*????      yorig = psGDC->hard_yorig; */

        hlf_barwdth     = (int)( (float)(PX(2)-PX(1)) * (((float)psGDC->bar_width/100.0)/2.0) );   /* used only for bars */
        hlf_hlccapwdth  = (int)( (float)(PX(2)-PX(1)) * (((float)psGDC->HLC_cap_width/100.0)/2.0) );
        }
    else
        {
            /* Flip axis, for horizontal bar charts */
        yorig = (float)( ( graphwidth +
                           vtitle_hgt +
                           vtics      +
                           vlabel_wth +
                           title_hgt  +
                           MAX(ltitle_hgt, rtitle_hgt) +
                           annote_hgt + 2) );
        if( psGDC->hard_size && psGDC->hard_yorig )
            yorig = psGDC->hard_yorig;
        xorig = (float)(  psGDC->imgwidth - (grapheight
                                             + ABS(xscl) * MIN(lowest,highest)
                                             + xdepth_3Dtotal + 2) );

        /* false_xorig is the actual origin to use for plotting values on H2BAR types
         * it is the x coord. of the separation line between the 2 sides
         */
/*        false_xorig = xorig + (do_h2bar? (highest-lowest)*ABS(xscl): 0); */
        if (do_h2bar)
        {
            if (highest > 0)
                false_xorig = xorig + (highest)*ABS(xscl);
            else  /* lowest < 0 */
                false_xorig = xorig + (-lowest)*ABS(xscl);

        }
        else
        {
            false_xorig = xorig + ABS(xscl) * MIN(lowest,highest);
        }

        hlf_barwdth     = (int)( (float)(PY(2)-PY(1)) * (((float)psGDC->bar_width/100.0)/2.0) );   /* used only for bars */
        hlf_hlccapwdth  = (int)( (float)(PY(2)-PY(1)) * (((float)psGDC->HLC_cap_width/100.0)/2.0) );
        }

    psGDC->hard_xorig = xorig;
    psGDC->hard_yorig = yorig;

    }
    /* scaled, sized, ready */


    /* ----- OK start the graphic ----- */
    if( (psGDC->hold_img & GDC_REUSE_IMAGE) &&
        psGDC->image != (void*)NULL )
        im = (gdImagePtr)psGDC->image;
    else
        im = gdImageCreate( psGDC->imgwidth, psGDC->imgheight );


    BGColor        = gdImageColorAllocate( im, l2gdcal(psGDC->BGColor) );
    LineColor      = clrallocate( im, psGDC->LineColor );
    PlotColor      = clrallocate( im, psGDC->PlotColor );
    GridColor      = clrallocate( im, psGDC->GridColor );
    if( do_vol )
      {
      VolColor     = clrallocate( im, psGDC->VolColor );
      for( i=0; i<psGDC->num_points; ++i )
        if( psGDC->ExtVolColor )
          ExtVolColor[i] = clrallocate( im, psGDC->ExtVolColor[i] );
        else
          ExtVolColor[i] = VolColor;
      }
/*  ArrowDColor    = gdImageColorAllocate( im, 0xFF,    0, 0 ); */
/*  ArrowUColor    = gdImageColorAllocate( im,    0, 0xFF, 0 ); */
    if( psGDC->annotation )
        AnnoteColor = clrallocate( im, psGDC->annotation->color );

    /* attempt to import optional background image */
    if( psGDC->BGImage )
        {
        FILE    *in = fopen(psGDC->BGImage, "rb");
        if( !in )
            {
                ; /* Cant load background image, drop it */
            }
        else
            {
            /* assume GIF */
            /* should determine type by file extension, option, ... */
            if( (bg_img = gdImageCreateFromGif(in)) )                 /* = */
                {
                int bgxpos = gdImageSX(bg_img)<psGDC->imgwidth?  psGDC->imgwidth/2 - gdImageSX(bg_img)/2:  0,
                    bgypos = gdImageSY(bg_img)<psGDC->imgheight? psGDC->imgheight/2 - gdImageSY(bg_img)/2: 0;


                if( gdImageSX(bg_img) > psGDC->imgwidth ||             /* resize only if too big */
                    gdImageSY(bg_img) > psGDC->imgheight )             /*  [and center] */
                    {
                    gdImageCopyResized( im, bg_img,             /* dst, src   */
                                        bgxpos, bgypos,         /* dstX, dstY */
                                        0, 0,                   /* srcX, srcY */
                                        psGDC->imgwidth, psGDC->imgheight,    /* dstW, dstH */
                                        psGDC->imgwidth, psGDC->imgheight );  /* srcW, srcH */
                    }
                else                                            /* just center */
                    gdImageCopy( im, bg_img,                    /* dst, src    */
                                 bgxpos, bgypos,                /* dstX, dstY  */
                                 0, 0,                          /* srcX, srcY  */
                                 psGDC->imgwidth, psGDC->imgheight );         /* W, H        */
                }
            fclose(in);
            }
        }

    /* Allocate colors */
    for( j=0; j<psGDC->num_sets; ++j )
        for( i=0; i<psGDC->num_points; ++i )
            if( psGDC->ExtColor )
                {
                unsigned long   ext_clr = *(psGDC->ExtColor+psGDC->num_points*j+i);

                _2D(ExtColor,j,i)            = clrallocate( im, ext_clr );
                if( threeD )
                    _2D(ExtColorShd,j,i)     = clrshdallocate( im, ext_clr );
                }
            else if( psGDC->SetColor )
                {
                int set_clr = psGDC->SetColor[j];
                _2D(ExtColor,j,i)     = clrallocate( im, set_clr );
                if( threeD )
                 _2D(ExtColorShd,j,i) = clrshdallocate( im, set_clr );
                }
            else
                {
                _2D(ExtColor,j,i)     = PlotColor;
                if( threeD )
                 _2D(ExtColorShd,j,i) = clrshdallocate( im, psGDC->PlotColor );
                }
            
    if( psGDC->transparent_bg )
        gdImageColorTransparent( im, BGColor );


    /* Create line styles, makes sense only for 2D LINE charts.
     * psGDC->LineStyles is either NULL (no styles), or an array of num_sets
     * style definition strings in which a space means transparent, and 
     * anything else means the pixel is turned on.
     *
     * Note that we cannot switch colors along a line using the colors
     * from ExtColor because this breaks the line pattern.  So we will
     * always use the color from ExtColor[j][0] for each line... this should be
     * OK for most cases... who needs multicolor lines anyways??? ;-)
     */
    if (psGDC->type == GDC_LINE ||
        psGDC->type ==  GDC_COMBO_LINE_BAR ||
        psGDC->type ==  GDC_COMBO_LINE_AREA||
        psGDC->type ==  GDC_COMBO_HLC_LINE ||
        psGDC->type ==  GDC_COMBO_LINE_HLC ||
        psGDC->type ==  GDC_3DCOMBO_HLC_LINE||
        psGDC->type ==  GDC_COMBO_LINE_LINE)
    {
        ALLOC_ARRAY(GDCLineStyle, LineStyles, psGDC->num_sets);

        for( j=0; j<psGDC->num_sets; ++j )
        {
            LineStyles[j].Style = NULL;
            LineStyles[j].StyleLength = 0;
            LineStyles[j].LineColor = _2D(ExtColor,j,0);

            if (psGDC->LineStyles)
            {
                /* Set to use line style */
                LineStyles[j].StyleLength = strlen(psGDC->LineStyles[j]);
                LineStyles[j].Style = (int*)malloc(LineStyles[j].StyleLength*sizeof(int));
                if (LineStyles[j].Style != NULL)
                {
                    /* Build the style array */
                    int k;
                    for(k=0; k<LineStyles[j].StyleLength; k++)
                    {
                        LineStyles[j].Style[k] = 
                            (psGDC->LineStyles[j][k]==' ')?gdTransparent:_2D(ExtColor,j,0);
                    }
                    LineStyles[j].LineColor = gdStyled;
                }
                else
                {
                    /* Allocation failed... just do as if there was no style... 
                     * the program will likely abort later anyways */
                    LineStyles[j].StyleLength = 0;
                }
            }
        }
    }/* If LINE chart type */

    {
        int tlen, numLines;
        int titlecolor = clrallocate( im, psGDC->TitleColor );

        if( psGDC->title )
        {
            GDC_cnt_nl( psGDC->title, &tlen );
            GDCImageStringNL( im,
                          &GDC_fontc[psGDC->title_size],
                          psGDC->imgwidth/2 - tlen*GDC_fontc[psGDC->title_size].w/2,
                          0,
                          psGDC->title,
                          titlecolor,
                          GDC_JUSTIFY_CENTER );
        }

        if(do_h2bar && psGDC->ltitle )
        {
            numLines = GDC_cnt_nl( psGDC->ltitle, &tlen );
            setno = psGDC->stack_type==GDC_STACK_DEPTH? psGDC->num_h2_sets:1;         /* backmost  */
            GDCImageStringNL( im,
                              &GDC_fontc[psGDC->title_size],
                              PXH2(-1, lowest+(highest-lowest)/2.0) - tlen*GDC_fontc[psGDC->title_size].w/2,
                              PY(psGDC->num_points+1) - 2 -numLines*GDC_fontc[psGDC->title_size].h,
                              psGDC->ltitle,
                              titlecolor,
                              GDC_JUSTIFY_CENTER );
            setno = 0;   /* reset to foremost */
        }
        if(do_h2bar && psGDC->rtitle )
        {
            numLines = GDC_cnt_nl( psGDC->rtitle, &tlen );
            setno = psGDC->stack_type==GDC_STACK_DEPTH? psGDC->num_h2_sets:1;         /* backmost  */
            GDCImageStringNL( im,
                              &GDC_fontc[psGDC->title_size],
                              PXH2(1, lowest+(highest-lowest)/2.0) - tlen*GDC_fontc[psGDC->title_size].w/2,
                              PY(psGDC->num_points+1) - 2 -numLines*GDC_fontc[psGDC->title_size].h,
                              psGDC->rtitle,
                              titlecolor,
                              GDC_JUSTIFY_CENTER );
            setno = 0;   /* reset to foremost */
        }
    }

    if( psGDC->xtitle )
        {
        int titlecolor = psGDC->XTitleColor==GDC_DFLTCOLOR? 
                            PlotColor: clrallocate( im, psGDC->XTitleColor );
        if ( ! flip_axis )
            gdImageString( im,
                           GDC_fontc[psGDC->xtitle_size].f,
                           psGDC->imgwidth/2 - strlen(psGDC->xtitle)*GDC_fontc[psGDC->xtitle_size].w/2,
                           psGDC->imgheight-GDC_fontc[psGDC->xtitle_size].h-1,
                           (unsigned char *)psGDC->xtitle,
                           titlecolor );
        else
            gdImageStringUp( im,
                             GDC_fontc[psGDC->xtitle_size].f,
                             0,
                             psGDC->imgheight/2 + strlen(psGDC->xtitle)*GDC_fontc[psGDC->xtitle_size].w/2,
                             (unsigned char *)psGDC->xtitle,
                             titlecolor );
        }



    /* ----- start drawing ----- */
    /* ----- backmost first - border, grid & labels ----- */
    /* if no grid, on 3D, border needs to handle it */
    if( !psGDC->grid && threeD &&
        ((psGDC->border == GDC_BORDER_ALL) || (psGDC->border & GDC_BORDER_X) || (psGDC->border & GDC_BORDER_Y)) )
    {
        int     x1, x2, y1, y2;

        x1 = PX(0);
        y1 = PY(lowest);

        setno = set_depth;
        x2 = PX(0);
        y2 = PY(lowest);

        gdImageLine( im, x1, y1, x2, y2, LineColor );     /* depth at origin */
        if( (psGDC->border == GDC_BORDER_ALL) || (psGDC->border & GDC_BORDER_X) )
            gdImageLine( im, x2, y2, PX(psGDC->num_points-1+(do_bar?2:0)), y2, LineColor );
        if( (psGDC->border == GDC_BORDER_ALL) || (psGDC->border & GDC_BORDER_Y) )
            gdImageLine( im, x2, PY(highest), x2, y2, LineColor );
        setno = 0;
    }
    if( psGDC->grid || psGDC->ticks || psGDC->yaxis )
        {   /* grid lines & y label(s) */
        float   tmp_y = lowest;
        int     labelcolor = psGDC->YLabelColor==GDC_DFLTCOLOR? 
                             LineColor: clrallocate( im, psGDC->YLabelColor );
        int     label2color = psGDC->YLabel2Color==GDC_DFLTCOLOR? 
                              VolColor: clrallocate( im, psGDC->YLabel2Color );
      int side;
      for(side=(do_h2bar?-1:1); side <=1 ; side+=2)
      {
        /* side=-1: bars to the left, only for H2BAR types
         * side=1: bars to the right, always present
         */  

        /* step from lowest to highest puting in labels and grid at interval points */
        /* since now "odd" intervals may be requested, try to step starting at 0,   */
        /* if lowest < 0 < highest                                                  */
        for( i=-1; i<=1; i+=2 )                                 /* -1, 1  */
            {
            int cur_loglabel=0;  /* Used for Log scale only */

            if( i == -1 )   if( lowest >= 0.0 )                 /*  all pos plotting */
                                continue;
                            else
                                tmp_y = MIN( 0, highest );      /*  step down to lowest */

            if( i == 1 )    if( highest <= 0.0 )                /*  all neg plotting */
                                continue;
                            else
                                tmp_y = MAX( 0, lowest );       /*  step up to highest */

            if (psGDC->yscale_type == GDC_SCALE_LOG && i==-1)
                continue;

/*
//          if( !(highest > 0 && lowest < 0) )                  // doesn't straddle 0
//              {
//              if( i == -1 )                                   // only do once: normal
//                  continue;
//              }
//          else
//              tmp_y = 0;
*/
            do  /* while( (tmp_y (+-)= ylbl_interval) < [highest,lowest] ) */
                {
                int     n, d, w;
                char    nmrtr[3+1], dmntr[3+1], whole[8];
                char    all_whole = ylbl_interval<1.0? FALSE: TRUE;

                char    *ylbl_str;

                if (psGDC->yscale_type == GDC_SCALE_LOG)
                    tmp_y = log_ylbl[cur_loglabel];

                ylbl_str = GDC_price_to_str( tmp_y,&n,&d,&w,
                                             do_ylbl_fractions? NULL: 
                                                           psGDC->ylabel_fmt );
                if( do_ylbl_fractions )
                    {
                    sprintf( nmrtr, "%d", n );
                    sprintf( dmntr, "%d", d );
                    sprintf( whole, "%d", w );
                    }

                if (psGDC->yscale_type == GDC_SCALE_LOG) 
                    tmp_y = log10(tmp_y);   /* Draw grid using log value */

                if( psGDC->grid || psGDC->ticks )
                    {
                    int x1, x2, x3, y1, y2, y3;
                    /* int  gridline_clr = tmp_y == 0.0? LineColor: GridColor; */
                                                                        /* tics */
                    x1 = flip_axis? PXH2(side, tmp_y) : PX(0);
                    y1 = flip_axis? PY(0) : PY(tmp_y);
                    if (psGDC->ticks)
                        gdImageLine( im, flip_axis? x1:(x1-2), flip_axis? y1:(y1+2), x1, y1, GridColor );
                    if (psGDC->grid)
                        {
                            setno = set_depth;
                            x2 = flip_axis? PXH2(side, tmp_y) : PX(0);          /* w/ new setno */
                            y2 = flip_axis? PY(0) : PY(tmp_y);
                            gdImageLine( im, x1, y1, x2, y2, GridColor );       /* depth for 3Ds */
                            x3 = flip_axis? x2: PX(psGDC->num_points-1+(do_bar?2:0));
                            y3 = flip_axis? PY(psGDC->num_points-1+(do_bar?2:0)):y2;
                            gdImageLine( im, x2, y2, x3, y3, GridColor );
                            setno = 0;                                          /* set back to foremost */
                        }
                    }
                if( psGDC->yaxis )
                    if( do_ylbl_fractions )
                        {
                        if( w || (!w && !n && !d) )
                            {
                            int totlen = 2 + strlen(whole)*GDC_fontc[psGDC->yaxisfont_size].w
                                                  + ( (!all_whole)?
                                                        (strlen(nmrtr)*GDC_fontc[psGDC->yaxisfont_size-1].w +
                                                         GDC_fontc[psGDC->yaxisfont_size].w                 +
                                                         strlen(nmrtr)*GDC_fontc[psGDC->yaxisfont_size-1].w) :
                                                        1 );
                            if ( !flip_axis )
                                gdImageString( im,
                                               GDC_fontc[psGDC->yaxisfont_size].f,
                                               PX(0)-totlen, PY(tmp_y)-GDC_fontc[psGDC->yaxisfont_size].h/2,
                                               (unsigned char *)whole,
                                               labelcolor );
                            else
                                gdImageStringUp( im,
                                               GDC_fontc[psGDC->yaxisfont_size].f,
                                               PXH2(side, tmp_y)-GDC_fontc[psGDC->yaxisfont_size].h/2, PY(0)+totlen, 
                                               (unsigned char *)whole,
                                               labelcolor );

                            }
                        if( n )
                            {
                            int offset;
                            offset = 2+strlen(nmrtr)*GDC_fontc[psGDC->yaxisfont_size-1].w
                                                  +GDC_fontc[psGDC->yaxisfont_size].w
                                                  +strlen(nmrtr)*GDC_fontc[psGDC->yaxisfont_size-1].w - 1;
                            if ( !flip_axis )
                                gdImageString( im,
                                           GDC_fontc[psGDC->yaxisfont_size-1].f,
                                           PX(0)-offset,
                                           PY(tmp_y)-GDC_fontc[psGDC->yaxisfont_size].h/2 + 1,
                                           (unsigned char *)nmrtr,
                                           labelcolor );
                            else
                                gdImageStringUp( im,
                                           GDC_fontc[psGDC->yaxisfont_size-1].f,
                                           PXH2(side, tmp_y)-GDC_fontc[psGDC->yaxisfont_size].h/2 - 1,
                                           PY(0)+offset,
                                           (unsigned char *)nmrtr,
                                           labelcolor );

                            offset = 2+GDC_fontc[psGDC->yaxisfont_size].w
                                                  +strlen(nmrtr)*GDC_fontc[psGDC->yaxisfont_size-1].w;
                            if ( !flip_axis )
                                gdImageString( im,
                                           GDC_fontc[psGDC->yaxisfont_size].f,
                                           PX(0)-offset,
                                           PY(tmp_y)-GDC_fontc[psGDC->yaxisfont_size].h/2,
                                           (unsigned char *)"/",
                                           labelcolor );
                            else
                                gdImageStringUp( im,
                                           GDC_fontc[psGDC->yaxisfont_size].f,
                                           PXH2(side, tmp_y)-GDC_fontc[psGDC->yaxisfont_size].h/2,
                                           PY(0)+offset,
                                           (unsigned char *)"/",
                                           labelcolor );

                            offset = 2 + strlen(nmrtr)*GDC_fontc[psGDC->yaxisfont_size-1].w + 2;

                            if ( !flip_axis )
                                gdImageString( im,
                                           GDC_fontc[psGDC->yaxisfont_size-1].f,
                                           PX(0)-offset,
                                           PY(tmp_y)-GDC_fontc[psGDC->yaxisfont_size].h/2 + 3,
                                           (unsigned char *)dmntr,
                                           labelcolor );
                            else
                                gdImageStringUp( im,
                                           GDC_fontc[psGDC->yaxisfont_size-1].f,
                                           PXH2(side, tmp_y)-GDC_fontc[psGDC->yaxisfont_size].h/2 + 3,
                                           PY(0)+offset,
                                           (unsigned char *)dmntr,
                                           labelcolor );
                            }
                        }
                    else
                    {
                        if ( !flip_axis )
                            gdImageString( im,
                                       GDC_fontc[psGDC->yaxisfont_size].f,
                                       PX(0)-2-strlen(ylbl_str)*GDC_fontc[psGDC->yaxisfont_size].w,
                                       PY(tmp_y)-GDC_fontc[psGDC->yaxisfont_size].h/2,
                                       (unsigned char *)ylbl_str,
                                       labelcolor );
                        else
                            gdImageStringUp( im,
                                       GDC_fontc[psGDC->yaxisfont_size].f,
                                       PXH2(side, tmp_y)-GDC_fontc[psGDC->yaxisfont_size].h/2,
                                       PY(0)+2+strlen(ylbl_str)*GDC_fontc[psGDC->yaxisfont_size].w,
                                       (unsigned char *)ylbl_str,
                                       labelcolor );
                    }

                if( do_vol && psGDC->yaxis2 )
                    {
                    char    vylbl[16];
                                                                                /* opposite of PV(y) */
                    sprintf( vylbl,
                             psGDC->ylabel2_fmt? psGDC->ylabel2_fmt: "%.0f",
                             ((float)(PY(tmp_y)+(setno*ydepth_3D)-vyorig))/vyscl );

                    setno = set_depth;
                    if (psGDC->ticks)
                        gdImageLine( im, PX(psGDC->num_points-1+(do_bar?2:0)), PY(tmp_y),
                                         PX(psGDC->num_points-1+(do_bar?2:0))+3, PY(tmp_y), GridColor );
                    if( atof(vylbl) == 0.0 )                                    /* rounding can cause -0 */
                        strcpy( vylbl, "0" );
                    gdImageString( im,
                                   GDC_fontc[psGDC->yaxisfont_size].f,
                                   PX(psGDC->num_points-1+(do_bar?2:0))+6,
                                   PY(tmp_y)-GDC_fontc[psGDC->yaxisfont_size].h/2,
                                   (unsigned char *)vylbl,
                                   label2color );
                    setno = 0;
                    }
                }
            while( ( psGDC->yscale_type == GDC_SCALE_LINEAR &&
                     (((i>0) && ((tmp_y += ylbl_interval) < highest)) ||
                      ((i<0) && ((tmp_y -= ylbl_interval) > lowest))    ) ) ||
                    (psGDC->yscale_type == GDC_SCALE_LOG && ++cur_loglabel < log_num_ylbl) );
            }
      }/*for side = -1 , 1 */

        /* catch last (bottom) grid line - specific to an "off" requested interval */
        if( psGDC->grid && threeD )
            {
            setno = set_depth;
            gdImageLine( im, flip_axis? PX(lowest):PX(0),
                             flip_axis? PY(0):PY(lowest), 
                             flip_axis? PX(lowest):PX(psGDC->num_points-1+(do_bar?2:0)),
                             flip_axis? PY(psGDC->num_points-1+(do_bar?2:0)):PY(lowest), 
                             GridColor );
            setno = 0;                                          /* set back to foremost */
            }

        /* vy axis title */
        if( do_vol && psGDC->ytitle2 )
            {
            int titlecolor = psGDC->YTitle2Color==GDC_DFLTCOLOR? 
                                VolColor: clrallocate( im, psGDC->YTitle2Color );
            gdImageStringUp( im,
                             GDC_fontc[psGDC->ytitle_size].f,
                             psGDC->imgwidth-(1+GDC_fontc[psGDC->ytitle_size].h),
                             (strlen(psGDC->ytitle2)*GDC_fontc[psGDC->ytitle_size].w)/2 +
                                psGDC->imgheight/2,
                             (unsigned char *)psGDC->ytitle2,
                             titlecolor );
            }

        /* y axis title */
        if( psGDC->yaxis && psGDC->ytitle )
            {
            int ytit_len = strlen(psGDC->ytitle)*GDC_fontc[psGDC->ytitle_size].w;
            int titlecolor = psGDC->YTitleColor==GDC_DFLTCOLOR? 
                                PlotColor: clrallocate( im, psGDC->YTitleColor );
            if ( ! flip_axis )
                gdImageStringUp( im,
                                 GDC_fontc[psGDC->ytitle_size].f,
                                 0,
                                 psGDC->imgheight/2 + ytit_len/2,
                                 (unsigned char *)psGDC->ytitle,
                                 titlecolor );
            else
                gdImageString( im,
                                 GDC_fontc[psGDC->ytitle_size].f,
                                 (do_h2bar? false_xorig:psGDC->imgwidth/2) - ytit_len/2,
                                 psGDC->imgheight - GDC_fontc[psGDC->ytitle_size].h-1,
                                 (unsigned char *)psGDC->ytitle,
                                 titlecolor );
            }
        }

    /* interviening set grids */
    /*  0 < setno < psGDC->num_sets   non-inclusive, they've already been covered */
    if( psGDC->grid && threeD )
        {
        for( setno=set_depth - 1;
             setno > 0;
             --setno )
            {
            gdImageLine( im, PX(0), PY(lowest), PX(0), PY(highest), GridColor );
            gdImageLine( im, PX(0), PY(lowest), PX(psGDC->num_points-1+(do_bar?2:0)), PY(lowest), GridColor );
            }
        setno = 0;
        }

    if( psGDC->yscale_type != GDC_SCALE_LOG &&
        ( psGDC->grid || psGDC->_0Shelf ) &&                           /* line color grid at 0 */
        ( (lowest < 0.0 && highest > 0.0) ||                           /* applies to linear scale only */ 
          ((lowest == 0.0 || highest == 0.0) && !(psGDC->border&GDC_BORDER_X) ) ) )
        {
        int x1, x2, y1, y2;
                                                                /* tics */
        x1 = PX(0);     y1 = PY(0);
        if (psGDC->ticks)
            gdImageLine( im, x1-2, y1, x1, y1, LineColor );
        setno = set_depth;
        x2 = PX(0);     y2 = PY(0);                             /* w/ new setno */
        gdImageLine( im, x1, y1, x2, y2, LineColor );           /* depth for 3Ds */
        gdImageLine( im, x2, y2, PX(psGDC->num_points-1+(do_bar?2:0)), y2, LineColor );
        setno = 0;                                              /* set back to foremost */
        }


    /* x ticks and xlables */
    if( psGDC->grid || psGDC->xaxis )
        {
        int     num_xlbls =                                     /* maximum x lables that'll fit */
                                                                /* each xlbl + avg due to num_lf_xlbls */
                    graphwidth /
                        ( (psGDC->xlabel_spacing==MAXSHORT?0:psGDC->xlabel_spacing)+GDC_fontc[psGDC->xaxisfont_size].h +
                          (num_lf_xlbls*(GDC_fontc[psGDC->xaxisfont_size].h-1))/psGDC->num_points );
        int     labelcolor = psGDC->XLabelColor==GDC_DFLTCOLOR? 
                             LineColor: clrallocate( im, psGDC->XLabelColor );

        for( i=0; i<psGDC->num_points+(do_bar?2:0); ++i )
            {
            int xi = do_bar? i-1: i;
            int x1, x2, y1, y2, xh, yh;                               /* ticks & grids */
#define DO_TICK(x,y)            if( psGDC->ticks )                                              \
                                    gdImageLine( im, x, y,                                      \
                                                flip_axis?((x)-2):(x), flip_axis?(y):((y)+2), GridColor ); \
                                else
#define DO_GRID(x1,y1,x2,y2)    if( psGDC->grid )                                               \
                                    {                                                           \
                                    gdImageLine( im, x1, y1, x2,  y2, GridColor ); /* depth */  \
                                    gdImageLine( im, x2, y2,                                    \
                                                 flip_axis?(xh):(x2), flip_axis?(y2):(yh), GridColor ); \
                                    }                                                           \
                                 else

            x1 = flip_axis? PX(lowest): PX(i);
            y1 = flip_axis? PY(i): PY(lowest);
            setno = set_depth;
            x2 = flip_axis? PX(lowest): PX(i);
            y2 = flip_axis? PY(i): PY(lowest);
            xh = flip_axis? PXH2( 1, highest):x2;
            yh = flip_axis? y2:PY(highest);
            setno = 0;                                              // reset to foremost

            if( i == 0 )                                            // catch 3D Y back corner
                DO_GRID(x1,y1,x2,y2);

            /* labeled points */
            if( (!psGDC->xlabel_ctl && ( (i%(1+psGDC->num_points/num_xlbls) == 0) ||  // # x labels are regulated
                                          num_xlbls >= psGDC->num_points       ||
                                          psGDC->xlabel_spacing == MAXSHORT ))
                ||
                (psGDC->xlabel_ctl && xi>=0 && *(psGDC->xlabel_ctl+xi)) )
                {
                DO_TICK(x1,y1);                                         // labeled points tick & grid
                DO_GRID(x1,y1,x2,y2);

                if( !do_bar || (i>0 && xi<psGDC->num_points) )
                    if( psGDC->xaxis && xlbl && xlbl[xi] && *(xlbl[xi]) )
                        {
                        /* waiting for GDCImageStringUpNL() */
#define LBX     GDC_fontc[psGDC->xaxisfont_size]
                        int     xlen = 0;
                        short   xstrs_num = GDC_cnt_nl( xlbl[xi], &xlen );
                        /* char    sub_xlbl[xlen+1]; */
                        DEF_ARRAY(char, sub_xlbl, xlen+1);
//                      int     xlbl_strt = -1+ PX((float)i+(float)(do_bar?((float)psGDC->num_points/(float)num_xlbls):0.0)) - (int)((float)(LBX.h-2)*((float)xstrs_num/2.0));
                        int     xlbl_strt = -1+ (flip_axis?PY(i):PX(i)) - (int)((float)(LBX.h-2)*((float)xstrs_num/2.0));

                        ALLOC_ARRAY(char, sub_xlbl, xlen+1);

                        xlen      = -1;
                        xstrs_num = -1;
                        j = -1;
                        do
                            {
                            ++j;
                            ++xlen;
                            sub_xlbl[xlen] = xlbl[xi][j];
                            if( xlbl[xi][j] == '\n' ||
                                xlbl[xi][j] == '\0' )
                                {
                                sub_xlbl[xlen] = '\0';
                                ++xstrs_num;
                                if (!flip_axis)
                                    gdImageStringUp( im,
                                                     LBX.f,
                                                     xlbl_strt + (LBX.h-1)*xstrs_num,
                                                     PY(lowest) + 2 + 1 + LBX.w*xlen,
                                                     sub_xlbl,
                                                     labelcolor );
                                else
                                    gdImageString( im,
                                                   LBX.f,
                                                   PX(lowest) - 2 - 1 - LBX.w*xlen,
                                                   xlbl_strt - (LBX.h-1)*xstrs_num,
                                                   sub_xlbl,
                                                   labelcolor );
                                xlen = -1;
                                }
                            } while( xlbl[xi][j] );
                        FREE_ARRAY(sub_xlbl);
#undef LBX
                        }
                }
            /* every point, on-point */
                if( i>0 )
                    {
                    if( psGDC->grid == GDC_TICK_POINTS )               /* --- GRID --- */
                        DO_GRID( x1, y1, x2, y2 );
                    else if( psGDC->grid > GDC_TICK_NONE )
                        {
                        int k;
                        int xt;
                        int prevx      = PX(i-1);
                        int intrv_dist = (x1-prevx)/(psGDC->grid+1);
                        DO_GRID( x1, y1, x2, y2 );
                        for( k=0,          xt=prevx + intrv_dist;
                             k<psGDC->grid && xt<x1;
                             ++k,          xt += intrv_dist )
                            DO_GRID( xt, y1, xt+xdepth_3Dtotal, y2 );
                        }

                    if( psGDC->ticks == GDC_TICK_POINTS )              /* --- TICKS --- */
                        DO_TICK(x1,y1);
                    else if( psGDC->ticks > GDC_TICK_NONE )
                        {
                        int k;
                        int xt;
                        int prevx=PX(i-1);
                        int intrv_dist = (x1-prevx)/(psGDC->ticks+1);
                        DO_TICK( x1, y1 );
                        for( k=0,           xt=prevx + intrv_dist;
                             k<psGDC->ticks && xt<x1;
                             ++k,           xt += intrv_dist )
                            DO_TICK( xt, y1 );
                        }
                    }
            }
        }

    /* ----- secondary data plotting (volume) ----- */
    /*  so that grid lines appear under vol */
    if( do_vol )
        {
        setno = set_depth;
        if( psGDC->type == GDC_COMBO_HLC_BAR    ||
            psGDC->type == GDC_COMBO_LINE_BAR   ||
            psGDC->type == GDC_3DCOMBO_LINE_BAR ||
            psGDC->type == GDC_3DCOMBO_HLC_BAR )
            {
            if( uvol[0] != GDC_NOVALUE )
                draw_3d_bar( im, PX(0), PX(0)+hlf_barwdth,
                                 PV(0), PV(uvol[0]),
                                 0, 0,
                                 ExtVolColor[0],
                                 ExtVolColor[0], 
                             flip_axis );
            for( i=1; i<psGDC->num_points-1; ++i )
                if( uvol[i] != GDC_NOVALUE )
                    draw_3d_bar( im, PX(i)-hlf_barwdth, PX(i)+hlf_barwdth,
                                     PV(0), PV(uvol[i]),
                                     0, 0,
                                     ExtVolColor[i],
                                     ExtVolColor[i],
                                 flip_axis );
            if( uvol[i] != GDC_NOVALUE )
                draw_3d_bar( im, PX(i)-hlf_barwdth, PX(i),
                                 PV(0), PV(uvol[i]),
                                 0, 0,
                                 ExtVolColor[i],
                                 ExtVolColor[i],
                             flip_axis);
            }
        else
        if( psGDC->type == GDC_COMBO_HLC_AREA   ||
            psGDC->type == GDC_COMBO_LINE_AREA  ||
            psGDC->type == GDC_3DCOMBO_LINE_AREA||
            psGDC->type == GDC_3DCOMBO_HLC_AREA )
            {
            for( i=1; i<psGDC->num_points; ++i )
                if( uvol[i-1] != GDC_NOVALUE && uvol[i] != GDC_NOVALUE )
                    draw_3d_area( im, PX(i-1), PX(i),
                                     PV(0), PV(uvol[i-1]), PV(uvol[i]),
                                     0, 0,
                                     ExtVolColor[i],
                                     ExtVolColor[i] );
            }
		else
        if( psGDC->type == GDC_COMBO_LINE_LINE ||
            psGDC->type == GDC_COMBO_HLC_LINE  ||
            psGDC->type == GDC_3DCOMBO_HLC_LINE||
            psGDC->type == GDC_3DCOMBO_LINE_LINE )
            {
            for( i=1; i<psGDC->num_points; ++i )
                if( uvol[i-1] != GDC_NOVALUE && uvol[i] != GDC_NOVALUE )
                    gdImageLine( im, PX(i-1), PV(uvol[i-1]),
                                     PX(i),   PV(uvol[i]),
                                     ExtVolColor[i] );
            }
        else
        if( psGDC->type == GDC_COMBO_LINE_HLC   ||
            psGDC->type == GDC_3DCOMBO_LINE_HLC )
            {
            int iMax,iVal;
            iMax = psGDC->num_points;
            iVal = iMax * 2;
            for( i=0; i<psGDC->num_points; ++i )
                if( uvol[i+iVal] != GDC_NOVALUE )
                {      /* all HLC have vert line */
                    if( uvol[i] != GDC_NOVALUE )
                        gdImageLine( im, PX(i), PV(uvol[i]),
                                     PX(i), PV(uvol[i+iVal]), ExtVolColor[i] );
                    if( uvol[i+iMax] != GDC_NOVALUE )
                        gdImageLine( im, PX(i), PV(uvol[i+iMax]),
                                     PX(i), PV(uvol[i+iVal]), ExtVolColor[i] );

                    if( psGDC->HLC_style & GDC_HLC_I_CAP )
                        {
                        if( uvol[i] != GDC_NOVALUE )
                            gdImageLine( im, PX(i)-hlf_hlccapwdth, 
                                         PV(uvol[i]),
                                         PX(i)+hlf_hlccapwdth, 
                                         PV(uvol[i]), ExtVolColor[i] );
                        if( uvol[i+iMax] != GDC_NOVALUE )
                            gdImageLine( im, PX(i)-hlf_hlccapwdth, 
                                         PV(uvol[i+iMax]),
                                         PX(i)+hlf_hlccapwdth, 
                                         PV(uvol[i+iMax]), ExtVolColor[i] );
                        }
                    if( psGDC->HLC_style & GDC_HLC_DIAMOND )
                        {
                        gdPoint         cd[4];

                        cd[0].x = PX(i)-hlf_hlccapwdth; 
                        cd[0].y = PV(uvol[i+iVal]);
                        cd[1].x = PX(i);    
                        cd[1].y = PV(uvol[i+iVal])+hlf_hlccapwdth;
                        cd[2].x = PX(i)+hlf_hlccapwdth;
                        cd[2].y = PV(uvol[i+iVal]);
                        cd[3].x = PX(i);
                        cd[3].y = PV(uvol[i+iVal])-hlf_hlccapwdth;
                        gdImageFilledPolygon( im, cd, 4, ExtVolColor[i] );
                        }
                    }
            for( i=1; i<psGDC->num_points; ++i )
                if(uvol[i+iVal-1]!=GDC_NOVALUE && uvol[i+iVal] != GDC_NOVALUE )
                    {
                    /* line from prev close */
                    /* CLOSE_CONNECTED and CONNECTING are mutually exclusive */
                    if( psGDC->HLC_style & GDC_HLC_CLOSE_CONNECTED )
                        gdImageLine( im, PX(i-1), PV(uvol[i+iVal-1]),
                                     PX(i), PV(uvol[i+iVal-1]),ExtVolColor[i]);
                    else
                    /* thin connecting line */
                        if( psGDC->HLC_style & GDC_HLC_CONNECTING )
                            gdImageLine( im, PX(i-1), PV(uvol[i+iVal-1]),
                                        PX(i),PV(uvol[i+iVal]),ExtVolColor[i]);
                    }
            }
        setno = 0;
        }       /* volume polys done */

    if( psGDC->annotation && threeD )      /* back half of annotation line */
        {
        int x1 = PX(psGDC->annotation->point+(do_bar?1:0)),
            y1 = PY(lowest);
        setno = set_depth;
        gdImageLine( im, x1, y1, PX(psGDC->annotation->point+(do_bar?1:0)), PY(lowest), AnnoteColor );
        gdImageLine( im, PX(psGDC->annotation->point+(do_bar?1:0)), PY(lowest),
                         PX(psGDC->annotation->point+(do_bar?1:0)), PY(highest)-2, AnnoteColor );
        setno = 0;
        }

    /* ---------- Draw the Val label at the top of the chart ---------- */
    if ( psGDC->bar_labels )
    {
        char    bar_lbl[64];
        int     new_barwdth, xmodif, side, setno;
        int     labelcolor = psGDC->XLabelColor==GDC_DFLTCOLOR? 
                             LineColor: clrallocate( im, psGDC->XLabelColor );

        if(!flip_axis)
            xmodif = -(GDC_fontc[psGDC->xaxisfont_size].h/2) + xdepth_3D;
        else
            xmodif = -(GDC_fontc[psGDC->xaxisfont_size].h/2) - ydepth_3D;

        if( psGDC->num_sets > 1 )
        {
            if( psGDC->stack_type == GDC_STACK_BESIDE )
            {
                new_barwdth = (int)( (float)hlf_barwdth / ((float)psGDC->num_sets/2.0) );
                xmodif -= (new_barwdth/2);
            }
            else
            {
                GDC_out_err( psGDC, psGDC->imgwidth, psGDC->imgheight,
                             img_fptr,
                             psGDC->BGColor, psGDC->LineColor,
                             "BarLabels not implemented for this chart type." );
                FREE_ARRAY(ExtVolColor);
                FREE_ARRAY(ExtColor);
                FREE_ARRAY(ExtColorShd);

                return 2;
            }
        }
        else
        {
            new_barwdth = hlf_barwdth;
        }

        for( setno=psGDC->num_sets-1; setno>=0; --setno )      /* back sets first   PX, PY depth */
        {
            for( i=0; i<psGDC->num_points; ++i )
                if( uvals[setno][i] != GDC_NOVALUE )
                {
                    sprintf( bar_lbl, psGDC->bar_label_fmt? 
                             psGDC->bar_label_fmt : " %.0f ", 
                             uvals[setno][i] );

                    if (!flip_axis)
                        gdImageStringUp( im,
                                         GDC_fontc[psGDC->xaxisfont_size].f,
                                         PX(i+(do_bar?1:0)) + xmodif -
                                             hlf_barwdth + new_barwdth*(setno+1),
                                         PY(uvals[setno][i])-
                                             ydepth_3D,
                                         bar_lbl,
                                         labelcolor );
                    else
                        for(side=(do_h2bar?-1:1); side<=1; side+=2)
                            gdImageString( im,
                                           GDC_fontc[psGDC->xaxisfont_size].f,
                                           PXH2(side, uvals[setno][i])+
                                               xdepth_3D,
                                           PY(i+(do_bar?1:0)) + xmodif -
                                               hlf_barwdth + new_barwdth*(setno+1),
                                           bar_lbl,
                                           labelcolor );
                }
        }
        setno = 0;
    }

    /* ---------- start plotting the data ---------- */
    switch( psGDC->type )
        {
        case GDC_3DBAR:                 /* depth, width, y interval need to allow for whitespace between bars */
        case GDC_BAR:
        /* --------- */
        switch( psGDC->stack_type )
            {
            case GDC_STACK_DEPTH:
            for( setno=psGDC->num_sets-1; setno>=0; --setno )      /* back sets first   PX, PY depth */
                for( i=0; i<psGDC->num_points; ++i )
                    if( uvals[setno][i] != GDC_NOVALUE )
                        draw_3d_bar( im, PX(i+(do_bar?1:0))-hlf_barwdth, PX(i+(do_bar?1:0))+hlf_barwdth,
                                         PY(bars_ybase), PY(uvals[setno][i]),
                                         xdepth_3D, ydepth_3D,
                                         _2D(ExtColor,setno,i),
                                         threeD? _2D(ExtColorShd,setno,i): _2D(ExtColor,setno,i), 
                                     flip_axis );
            setno = 0;
            break;

            case GDC_STACK_LAYER:
                {
                /*float lasty[num_points]; */
                DEF_ARRAY(float, lasty, psGDC->num_points);
                ALLOC_ARRAY(float, lasty, psGDC->num_points);
                j = 0;
/*
//              for( i=0; i<psGDC->num_points; ++i )
//                  if( uvals[j][i] != GDC_NOVALUE )
//                      {
//                      lasty[i] = uvals[j][i];
//                      draw_3d_bar( im, PX(i+(do_bar?1:0))-hlf_barwdth, PX(i+(do_bar?1:0))+hlf_barwdth,
//                                       PY(0), PY(uvals[j][i]),
//                                       xdepth_3D, ydepth_3D,
//                                       ExtColor[j][i],
//                                       threeD? ExtColorShd[j][i]: ExtColor[j][i] );
//                      }
*/
                for( i=0; i<psGDC->num_points; ++i )
                    {
                    float       lasty_pos = 0.0;
                    float       lasty_neg = 0.0;
                    int         k;
                    /*struct BS barset[psGDC->num_sets]; */
                    DEF_ARRAY(struct BS, barset, psGDC->num_sets);
                    ALLOC_ARRAY(struct BS, barset, psGDC->num_sets);

                    for( j=0, k=0; j<psGDC->num_sets; ++j )
                        {
                        if( uvals[j][i] != GDC_NOVALUE )
                            {
                            if( uvals[j][i] < 0.0 )
                                {
                                barset[k].y1 = lasty_neg;
                                barset[k].y2 = uvals[j][i] + lasty_neg;
                                lasty_neg    = barset[k].y2;
                                }
                            else
                                {
                                barset[k].y1 = lasty_pos;
                                barset[k].y2 = uvals[j][i] + lasty_pos;
                                lasty_pos    = barset[k].y2;
                                }
                            barset[k].clr   = _2D(ExtColor,j,i);
                            barset[k].shclr = threeD? _2D(ExtColorShd,j,i): _2D(ExtColor,j,i);
                            ++k;
                            }
                        }
                    qsort( barset, k, sizeof(struct BS), barcmpr );

                    for( j=0; j<k; ++j )
                        {
                        draw_3d_bar( im,
                                     PX(i+(do_bar?1:0))-hlf_barwdth, PX(i+(do_bar?1:0))+hlf_barwdth,
                                     PY(barset[j].y1), PY(barset[j].y2),
                                     xdepth_3D, ydepth_3D,
                                     barset[j].clr,
                                     barset[j].shclr,
                                     flip_axis);
                        }
                    FREE_ARRAY(barset);
                    }
                FREE_ARRAY(lasty);
                }
                break;

            case GDC_STACK_BESIDE:
                {                                               /* h/.5, h/1, h/1.5, h/2, ... */
                int new_barwdth = (int)( (float)hlf_barwdth / ((float)psGDC->num_sets/2.0) );
                for( i=0; i<psGDC->num_points; ++i )
                    for( j=0; j<psGDC->num_sets; ++j )
                        if( uvals[j][i] != GDC_NOVALUE )
                            draw_3d_bar( im, PX(i+(do_bar?1:0))-hlf_barwdth+new_barwdth*j+1,
                                             PX(i+(do_bar?1:0))-hlf_barwdth+new_barwdth*(j+1),
                                             PY(bars_ybase), PY(uvals[j][i]),
                                             xdepth_3D, ydepth_3D,
                                             _2D(ExtColor,j,i),
                                             threeD? _2D(ExtColorShd,j,i): _2D(ExtColor,j,i),
                                         flip_axis);
                    }
                break;
            }
            break;

        case GDC_3DFLOATINGBAR:
        case GDC_FLOATINGBAR:
          /* --------- */
          switch( psGDC->stack_type )
            {
            case GDC_STACK_DEPTH:
                for( setno=psGDC->num_groups-1; setno>=0; --setno ) // back sets first   PX, PY depth
                    for( i=0; i<psGDC->num_points; ++i )
                        if( uvals[0+setno*2][i] != GDC_NOVALUE &&
                            uvals[1+setno*2][i] != GDC_NOVALUE &&
                            uvals[1+setno*2][i] > uvals[0+setno*2][i] )
                            draw_3d_bar( im, PX(i+(do_bar?1:0))-hlf_barwdth, PX(i+(do_bar?1:0))+hlf_barwdth,
                                         PY(uvals[0+setno*2][i]), PY(uvals[1+setno*2][i]),
                                         xdepth_3D, ydepth_3D,
                                         _2D(ExtColor,setno,i),
                                         threeD? _2D(ExtColorShd,setno,i): _2D(ExtColor,setno,i), flip_axis );
                setno = 0;
                break;

            case GDC_STACK_BESIDE:
                {                                               // h/.5, h/1, h/1.5, h/2, ...
                int new_barwdth = (int)( (float)hlf_barwdth / ((float)psGDC->num_groups/2.0) );
                for( i=0; i<psGDC->num_points; ++i )
                    for( j=0; j<psGDC->num_groups; ++j )
                        if( uvals[0+j*2][i] != GDC_NOVALUE &&
                            uvals[1+j*2][i] != GDC_NOVALUE &&
                            uvals[1+j*2][i] > uvals[0+j*2][i] )
                            draw_3d_bar( im, PX(i+(do_bar?1:0))-hlf_barwdth+new_barwdth*j+1,
                                             PX(i+(do_bar?1:0))-hlf_barwdth+new_barwdth*(j+1),
                                             PY(uvals[0+j*2][i]), PY(uvals[1+j*2][i]),
                                             xdepth_3D, ydepth_3D,
                                             _2D(ExtColor,j,i),
                                             threeD? _2D(ExtColorShd,j,i): _2D(ExtColor,j,i), flip_axis );
                }
                break;
            }
            break;

        case GDC_3DHBAR:                 /* depth, width, y interval need to allow for whitespace between bars */
        case GDC_HBAR:
        case GDC_3DH2BAR: 
        case GDC_H2BAR:
        /* --------- */
        {
        int side;
        for(side=(do_h2bar?-1:1); side <=1 ; side+=2)
        {
          /* side=-1: bars to the left, only for H2BAR types
           * side=1: bars to the right, always present
           */  
          switch( psGDC->stack_type )
            {
            case GDC_STACK_DEPTH:
            for( setno=psGDC->num_groups-1; setno>=0; --setno )      /* back sets first   PX, PY depth */
            {
                int h2set = (do_h2bar?(side==-1?H2_LEFTSET:H2_RIGHTSET):0)+(setno*(do_h2bar?2:1));
                for( i=0; i<psGDC->num_points; ++i )
                    if( uvals[h2set][i] != GDC_NOVALUE )
                        draw_3d_bar( im, PXH2(side, bars_ybase), PXH2(side, uvals[h2set][i]),
                                         PY(i+(do_bar?1:0))-hlf_barwdth, PY(i+(do_bar?1:0))+hlf_barwdth,
                                         xdepth_3D, ydepth_3D,
                                         _2D(ExtColor,h2set,i),
                                         threeD? _2D(ExtColorShd,h2set,i): _2D(ExtColor,h2set,i),
                                     flip_axis);

                setno = 0;
            }
            break;

            case GDC_STACK_LAYER:
                {
                int h2set; 
                /*float lasty[num_points]; */
                DEF_ARRAY(float, lasty, psGDC->num_points);
                ALLOC_ARRAY(float, lasty, psGDC->num_points);
                j = 0;
                h2set = (do_h2bar?(side==-1?H2_LEFTSET:H2_RIGHTSET):0)+(j*(do_h2bar?2:1));
/*
//              for( i=0; i<psGDC->num_points; ++i )
//                  if( uvals[j][i] != GDC_NOVALUE )
//                      {
//                      lasty[i] = uvals[j][i];
//                      draw_3d_bar( im, PY(i+(do_bar?1:0))-hlf_barwdth, PY(i+(do_bar?1:0))+hlf_barwdth,
//                                       PX(0), PX(uvals[j][i]),
//                                       xdepth_3D, ydepth_3D,
//                                       ExtColor[j][i],
//                                       threeD? ExtColorShd[j][i]: ExtColor[j][i] );
//                      }
*/
                for( i=0; i<psGDC->num_points; ++i )
                    {
                    float       lasty_pos = 0.0;
                    float       lasty_neg = 0.0;
                    int         k;
                    /*struct BS barset[psGDC->num_sets]; */
                    DEF_ARRAY(struct BS, barset, psGDC->num_groups);
                    ALLOC_ARRAY(struct BS, barset, psGDC->num_groups);

                    for( j=0, k=0; j<psGDC->num_groups; ++j )
                        {
                        h2set =(do_h2bar?(side==-1?H2_LEFTSET:H2_RIGHTSET):0)+(j*(do_h2bar?2:1));

                        if( uvals[h2set][i] != GDC_NOVALUE )
                            {
                            if( uvals[h2set][i] < 0.0 )
                                {
                                barset[k].y1 = lasty_neg;
                                barset[k].y2 = uvals[h2set][i] + lasty_neg;
                                lasty_neg    = barset[k].y2;
                                }
                            else
                                {
                                barset[k].y1 = lasty_pos;
                                barset[k].y2 = uvals[h2set][i] + lasty_pos;
                                lasty_pos    = barset[k].y2;
                                }
                            barset[k].clr   = _2D(ExtColor,h2set,i);
                            barset[k].shclr = threeD? _2D(ExtColorShd,h2set,i): _2D(ExtColor,h2set,i);
                            ++k;
                            }
                        }
                    qsort( barset, k, sizeof(struct BS), barcmpr );

                    for( j=0; j<k; ++j )
                        {
                        h2set = (do_h2bar?(side==-1?H2_LEFTSET:H2_RIGHTSET):0)+(j*(do_h2bar?2:1));
                        draw_3d_bar( im,
                                     PXH2(side, barset[h2set].y1), PXH2(side, barset[h2set].y2),
                                     PY(i+(do_bar?1:0))-hlf_barwdth, PY(i+(do_bar?1:0))+hlf_barwdth,
                                     xdepth_3D, ydepth_3D,
                                     barset[j].clr,
                                     barset[j].shclr,
                                     flip_axis);
                        }
                    FREE_ARRAY(barset);
                    }
                FREE_ARRAY(lasty);
                }
                break;
            case GDC_STACK_BESIDE:
                {                                               /* h/.5, h/1, h/1.5, h/2, ... */
                int new_barwdth = (int)( (float)hlf_barwdth / ((float)psGDC->num_groups/2.0) );
                for( i=0; i<psGDC->num_points; ++i )
                    for( j=0; j<psGDC->num_groups; ++j )
                    {
                        int h2set = (do_h2bar?(side==-1?H2_LEFTSET:H2_RIGHTSET):0)+(j*(do_h2bar?2:1));
                        if( uvals[h2set][i] != GDC_NOVALUE )
                            draw_3d_bar( im, PXH2(side, bars_ybase), PXH2(side, uvals[h2set][i]),
                                             PY(i+(do_bar?1:0))-hlf_barwdth+new_barwdth*j+1,
                                             PY(i+(do_bar?1:0))-hlf_barwdth+new_barwdth*(j+1),
                                             xdepth_3D, ydepth_3D,
                                             _2D(ExtColor,h2set,i),
                                             threeD? _2D(ExtColorShd,h2set,i): _2D(ExtColor,h2set,i),
                                         flip_axis);
                    }
                }
                break;
            }

            /* Vertical separation line between the 2 sides for H2BAR types */
            if( do_h2bar && psGDC->grid && side == -1)
            {
                int x1, x2, y1, y2;

                setno = psGDC->stack_type==GDC_STACK_DEPTH? psGDC->num_h2_sets: 1; /* backmost */
                gdImageLine( im, PXH2(1, lowest), PY(0),
                             PXH2(1, lowest), PY(psGDC->num_points+1), LineColor); 

                /* depth perspective */
                for( i=0; i<psGDC->num_points+2; ++i )
                {
                    setno = 0;                                         /* reset to foremost */
                    x1 = PXH2(1, lowest);
                    y1 = PY(i);
                    setno = psGDC->stack_type==GDC_STACK_DEPTH? psGDC->num_h2_sets: 1; /* backmost */
                    x2 = PXH2(1, lowest);
                    y2 = PY(i);

                    gdImageLine( im, x1, y1, x2,  y2, 
                                 (i==0||i==psGDC->num_points+1)? LineColor:GridColor ); 
                    setno = 0;                                          /* reset to foremost */
                }
                
            }
            else if( do_h2bar && psGDC->grid && side == 1)
            {
                /* foremost vertical line drawn last, on top of all bars */
                setno = 0;                                         /* reset to foremost */
                gdImageLine( im, PXH2(1, lowest), PY(0),
                             PXH2(1, lowest), PY(psGDC->num_points+1), LineColor); 
            }

          }/* for side */

        }
        break;

        case GDC_LINE:
        case GDC_COMBO_LINE_BAR:
        case GDC_COMBO_LINE_AREA:
        case GDC_COMBO_LINE_LINE:
        case GDC_COMBO_LINE_HLC:
            for( j=psGDC->num_sets-1; j>=0; --j )
            {
                if (LineStyles[j].LineColor == gdStyled)
                    gdImageSetStyle( im, LineStyles[j].Style, LineStyles[j].StyleLength);

                for( i=1; i<psGDC->num_points; ++i )
                {
                    if( uvals[j][i-1] != GDC_NOVALUE && uvals[j][i] != GDC_NOVALUE )
                        {
                        GDCImageLine( im, PX(i-1), PY(uvals[j][i-1]), PX(i), PY(uvals[j][i]), LineStyles[j].LineColor );
                        }
                    else
                        {
                        if( uvals[j][i-1] != GDC_NOVALUE )
                            gdImageSetPixel( im, PX(i-1), PY(uvals[j][i-1]), _2D(ExtColor,j,i) );
                        if( uvals[j][i] != GDC_NOVALUE )
                            gdImageSetPixel( im, PX(i), PY(uvals[j][i]), _2D(ExtColor,j,i) );
                        }
                }
            }
            break;

        case GDC_3DLINE:
        case GDC_3DCOMBO_LINE_BAR:
        case GDC_3DCOMBO_LINE_AREA:
	case GDC_3DCOMBO_LINE_LINE:
        case GDC_3DCOMBO_LINE_HLC:
            {
            /*int   y1[psGDC->num_sets], */
            /*  y2[num_sets]; */
            DEF_ARRAY(int, y1, psGDC->num_sets);
            DEF_ARRAY(int, y2, psGDC->num_sets);
            ALLOC_ARRAY(int, y1, psGDC->num_sets);
            ALLOC_ARRAY(int, y2, psGDC->num_sets);

            for( i=1; i<psGDC->num_points; ++i )
                {
                if( psGDC->stack_type == GDC_STACK_DEPTH )
                    {
                    for( j=psGDC->num_sets-1; j>=0; --j )
                        if( uvals[j][i-1] != GDC_NOVALUE &&
                            uvals[j][i]   != GDC_NOVALUE )
                            {
                            setno = j;
                            y1[j] = PY(uvals[j][i-1]);
                            y2[j] = PY(uvals[j][i]);

                            draw_3d_line( im,
                                          PY(0),
                                          PX(i-1), PX(i), 
                                          &(y1[j]), &(y2[j]),
                                          xdepth_3D, ydepth_3D,
                                          1,
                                          &(_2D(ExtColor,j,i)),
                                          &(_2D(ExtColorShd,j,i)) );
                            setno = 0;
                            }
                    }
                else
                if( psGDC->stack_type == GDC_STACK_BESIDE ||
                    psGDC->stack_type == GDC_STACK_SUM )           /* all same plane */
                    {
                    int     set;
                    float   usey1 = 0.0,
                            usey2 = 0.0;
                    /*int       clr[num_sets], */
                    /*      clrshd[num_sets]; */
                    DEF_ARRAY(int, clr, psGDC->num_sets);
                    DEF_ARRAY(int, clrshd, psGDC->num_sets);
                    ALLOC_ARRAY(int, clr, psGDC->num_sets);
                    ALLOC_ARRAY(int, clrshd, psGDC->num_sets);

                    for( j=0,set=0; j<psGDC->num_sets; ++j )
                        if( uvals[j][i-1] != GDC_NOVALUE &&
                            uvals[j][i]   != GDC_NOVALUE )
                            {
                            if( psGDC->stack_type == GDC_STACK_SUM )
                                {
                                usey1 += uvals[j][i-1];
                                usey2 += uvals[j][i];
                                }
                            else
                                {
                                usey1 = uvals[j][i-1];
                                usey2 = uvals[j][i];
                                }
                            y1[set]     = PY(usey1);
                            y2[set]     = PY(usey2);
                            clr[set]    = _2D(ExtColor,j,i);
                            clrshd[set] = _2D(ExtColorShd,j,i); /* fred */
                            ++set;
                            }
                    draw_3d_line( im,
                          PY(0),
                          PX(i-1), PX(i), 
                          y1, y2,
                          xdepth_3D, ydepth_3D,
                          set,
                          clr,
                          clrshd );
                    FREE_ARRAY(clr);
                    FREE_ARRAY(clrshd);
                    }
                }
            FREE_ARRAY(y1);
            FREE_ARRAY(y2);
            }
            break;

        case GDC_AREA:
        case GDC_3DAREA:
          switch( psGDC->stack_type )
            {
            case GDC_STACK_SUM:
                {
                /*float lasty[num_points]; */
                DEF_ARRAY(float, lasty, psGDC->num_points);
                ALLOC_ARRAY(float, lasty, psGDC->num_points);
                j = 0;
                for( i=1; i<psGDC->num_points; ++i )
                    if( uvals[j][i] != GDC_NOVALUE )
                        {
                        lasty[i] = uvals[j][i];
                        if( uvals[j][i-1] != GDC_NOVALUE )
                            draw_3d_area( im, PX(i-1), PX(i),
                                             PY(0), PY(uvals[j][i-1]), PY(uvals[j][i]),
                                             xdepth_3D, ydepth_3D,
                                             _2D(ExtColor,j,i),
                                             threeD? _2D(ExtColorShd,j,i): _2D(ExtColor,j,i) );
                        }
                for( j=1; j<psGDC->num_sets; ++j )
                    for( i=1; i<psGDC->num_points; ++i )
                        if( uvals[j][i] != GDC_NOVALUE && uvals[j][i-1] != GDC_NOVALUE )
                            {
                            draw_3d_area( im, PX(i-1), PX(i),
                                             PY(lasty[i]), PY(lasty[i-1]+uvals[j][i-1]), PY(lasty[i]+uvals[j][i]),
                                             xdepth_3D, ydepth_3D,
                                             _2D(ExtColor,j,i),
                                             threeD? _2D(ExtColorShd,j,i): _2D(ExtColor,j,i) );
                            lasty[i] += uvals[j][i];
                            }
                FREE_ARRAY(lasty);
                }
                break;

            case GDC_STACK_BESIDE:                              /* behind w/o depth */
                for( j=psGDC->num_sets-1; j>=0; --j )                  /* back sets 1st  (setno = 0) */
                    for( i=1; i<psGDC->num_points; ++i )
                        if( uvals[j][i-1] != GDC_NOVALUE && uvals[j][i] != GDC_NOVALUE )
                            draw_3d_area( im, PX(i-1), PX(i),
                                             PY(0), PY(uvals[j][i-1]), PY(uvals[j][i]),
                                             xdepth_3D, ydepth_3D,
                                             _2D(ExtColor,j,i),
                                             threeD? _2D(ExtColorShd,j,i): _2D(ExtColor,j,i) );
                break;

            case GDC_STACK_DEPTH:
            default:
                for( setno=psGDC->num_sets-1; setno>=0; --setno )      /* back sets first   PX, PY depth */
                    for( i=1; i<psGDC->num_points; ++i )
                        if( uvals[setno][i-1] != GDC_NOVALUE && uvals[setno][i] != GDC_NOVALUE )
                            draw_3d_area( im, PX(i-1), PX(i),
                                             PY(0), PY(uvals[setno][i-1]), PY(uvals[setno][i]),
                                             xdepth_3D, ydepth_3D,
                                             _2D(ExtColor,setno,i),
                                             threeD? _2D(ExtColorShd,setno,i): _2D(ExtColor,setno,i) );
                setno = 0;
            }
            break;

        case GDC_3DHILOCLOSE:
        case GDC_3DCOMBO_HLC_BAR:
        case GDC_3DCOMBO_HLC_AREA:
        case GDC_3DCOMBO_HLC_LINE:
            {
            gdPoint     poly[4];
            for( j=psGDC->num_groups-1; j>=0; --j )
             {
             for( i=1; i<psGDC->num_points+1; ++i )
                 if( uvals[CLOSESET+j*3][i-1] != GDC_NOVALUE )
                     {
                     if( (psGDC->HLC_style & GDC_HLC_I_CAP) &&         /* bottom half of 'I' */
                         uvals[LOWSET+j*3][i-1] != GDC_NOVALUE )
                         {
                         SET_3D_POLY( poly, PX(i-1)-hlf_hlccapwdth, PX(i-1)+hlf_hlccapwdth,
                                            PY(uvals[LOWSET+j*3][i-1]), PY(uvals[LOWSET+j*3][i-1]),
                                            xdepth_3D, ydepth_3D );
                         gdImageFilledPolygon( im, poly, 4, _2D(ExtColor,LOWSET+j*3,i-1) );
                         gdImagePolygon( im, poly, 4, _2D(ExtColorShd,LOWSET+j*3,i-1) );
                         }
                                                                     /* all HLC have vert line */
                     if( uvals[LOWSET+j*3][i-1] != GDC_NOVALUE )
                         {                                          /* bottom 'half' */
                         SET_3D_POLY( poly, PX(i-1), PX(i-1),
                                            PY(uvals[LOWSET+j*3][i-1]), PY(uvals[CLOSESET+j*3][i-1]),
                                            xdepth_3D, ydepth_3D );
                         gdImageFilledPolygon( im, poly, 4, _2D(ExtColor,LOWSET+j*3,i-1) );
                         gdImagePolygon( im, poly, 4, _2D(ExtColorShd,LOWSET+j*3,i-1) );
                         }
                     if( uvals[HIGHSET+j*3][i-1] != GDC_NOVALUE )
                         {                                          /* top 'half' */
                         SET_3D_POLY( poly, PX(i-1), PX(i-1),
                                            PY(uvals[CLOSESET+j*3][i-1]), PY(uvals[HIGHSET+j*3][i-1]),
                                            xdepth_3D, ydepth_3D );
                         gdImageFilledPolygon( im, poly, 4, _2D(ExtColor,HIGHSET+j*3,i-1) );
                         gdImagePolygon( im, poly, 4, _2D(ExtColorShd,HIGHSET+j*3,i-1) );
                         }
                                                                    /* line at close */
                     gdImageLine( im, PX(i-1),           PY(uvals[CLOSESET+j*3][i-1]),
                                      PX(i-1)+xdepth_3D, PY(uvals[CLOSESET+j*3][i-1])-ydepth_3D,
                                      _2D(ExtColorShd,CLOSESET+j*3,i-1) );
                                                                 /* top half 'I' */
                     if( !( (psGDC->HLC_style & GDC_HLC_DIAMOND) &&
                            (PY(uvals[HIGHSET+j*3][i-1]) > PY(uvals[CLOSESET+j*3][i-1])-hlf_hlccapwdth) ) &&
                         uvals[HIGHSET+j*3][i-1] != GDC_NOVALUE )
                         if( psGDC->HLC_style & GDC_HLC_I_CAP )
                             {
                             SET_3D_POLY( poly, PX(i-1)-hlf_hlccapwdth, PX(i-1)+hlf_hlccapwdth,
                                                PY(uvals[HIGHSET+j*3][i-1]), PY(uvals[HIGHSET+j*3][i-1]),
                                                xdepth_3D, ydepth_3D );
                             gdImageFilledPolygon( im, poly, 4, _2D(ExtColor,HIGHSET+j*3,i-1) );
                             gdImagePolygon( im, poly, 4, _2D(ExtColorShd,HIGHSET+j*3,i-1) );
                             }

                     if( i < psGDC->num_points &&
                         uvals[CLOSESET+j*3][i] != GDC_NOVALUE )
                         {
                         if( psGDC->HLC_style & GDC_HLC_CLOSE_CONNECTED )  /* line from prev close */
                             {
                             SET_3D_POLY( poly, PX(i-1), PX(i),
                                                PY(uvals[CLOSESET+j*3][i-1]), PY(uvals[CLOSESET+j*3][i-1]),
                                                xdepth_3D, ydepth_3D );
                             gdImageFilledPolygon( im, poly, 4, _2D(ExtColor,CLOSESET+j*3,i) );
                             gdImagePolygon( im, poly, 4, _2D(ExtColorShd,CLOSESET+j*3,i) );
                             }
                         else   /* CLOSE_CONNECTED and CONNECTING are mutually exclusive */
                         if( psGDC->HLC_style & GDC_HLC_CONNECTING )   /* thin connecting line */
                             {
                             int    y1 = PY(uvals[CLOSESET+j*3][i-1]),
                                 y2 = PY(uvals[CLOSESET+j*3][i]);
                             draw_3d_line( im,
                                           PY(0),
                                           PX(i-1), PX(i),
                                           &y1, &y2,                    /* rem only 1 set */
                                           xdepth_3D, ydepth_3D,
                                           1,
                                           &(_2D(ExtColor,CLOSESET+j*3,i)),
                                           &(_2D(ExtColorShd,CLOSESET+j*3,i)) );
                                                                     /* edge font of it */
                             gdImageLine( im, PX(i-1), PY(uvals[CLOSESET+j*3][i-1]),
                                              PX(i), PY(uvals[CLOSESET+j*3][i]),
                                              _2D(ExtColorShd,CLOSESET+j*3,i) );
                             }
                                                                     /* top half 'I' again */
                         if( PY(uvals[CLOSESET+j*3][i-1]) <= PY(uvals[CLOSESET+j*3][i]) &&
                             uvals[HIGHSET+j*3][i-1] != GDC_NOVALUE  )
                             if( psGDC->HLC_style & GDC_HLC_I_CAP )
                                 {
                                 SET_3D_POLY( poly, PX(i-1)-hlf_hlccapwdth, PX(i-1)+hlf_hlccapwdth,
                                                    PY(uvals[HIGHSET+j*3][i-1]), PY(uvals[HIGHSET+j*3][i-1]),
                                                    xdepth_3D, ydepth_3D );
                                 gdImageFilledPolygon( im, poly, 4, _2D(ExtColor,HIGHSET+j*3,i-1) );
                                 gdImagePolygon( im, poly, 4, _2D(ExtColorShd,HIGHSET+j*3,i-1) );
                                 }
                         }
                     if( psGDC->HLC_style & GDC_HLC_DIAMOND )
                         {                                  /* front */
                         poly[0].x = PX(i-1)-hlf_hlccapwdth;
                          poly[0].y = PY(uvals[CLOSESET+j*3][i-1]);
                         poly[1].x = PX(i-1);
                          poly[1].y = PY(uvals[CLOSESET+j*3][i-1])+hlf_hlccapwdth;
                         poly[2].x = PX(i-1)+hlf_hlccapwdth;
                          poly[2].y = PY(uvals[CLOSESET+j*3][i-1]);
                         poly[3].x = PX(i-1);
                          poly[3].y = PY(uvals[CLOSESET+j*3][i-1])-hlf_hlccapwdth;
                         gdImageFilledPolygon( im, poly, 4, _2D(ExtColor,CLOSESET+j*3,i-1) );
                         gdImagePolygon( im, poly, 4, _2D(ExtColorShd,CLOSESET+j*3,i-1) );
                                                             /* bottom side */
                         SET_3D_POLY( poly, PX(i-1), PX(i-1)+hlf_hlccapwdth,
                                            PY(uvals[CLOSESET+j*3][i-1])+hlf_hlccapwdth,
                                                     PY(uvals[CLOSESET+j*3][i-1]),
                                            xdepth_3D, ydepth_3D );
                         gdImageFilledPolygon( im, poly, 4, _2D(ExtColorShd,CLOSESET+j*3,i-1) );
                         /* gdImagePolygon( im, poly, 4, ExtColor[CLOSESET+j*3][i-1] ); */
                                                             /* top side */
                         SET_3D_POLY( poly, PX(i-1), PX(i-1)+hlf_hlccapwdth,
                                            PY(uvals[CLOSESET+j*3][i-1])-hlf_hlccapwdth,
                                                     PY(uvals[CLOSESET+j*3][i-1]),
                                            xdepth_3D, ydepth_3D );
                         gdImageFilledPolygon( im, poly, 4, _2D(ExtColor,CLOSESET+j*3,i-1) );
                         gdImagePolygon( im, poly, 4, _2D(ExtColorShd,CLOSESET+j*3,i-1) );
                         }
                     }
             }
            }
            break;

        case GDC_HILOCLOSE:
        case GDC_COMBO_HLC_BAR:
        case GDC_COMBO_HLC_AREA:
        case GDC_COMBO_HLC_LINE:
            for( j=psGDC->num_groups-1; j>=0; --j )
                {
                for( i=0; i<psGDC->num_points; ++i )
                    if( uvals[CLOSESET+j*3][i] != GDC_NOVALUE )
                        {                                           /* all HLC have vert line */
                        if( uvals[LOWSET+j*3][i] != GDC_NOVALUE )
                            gdImageLine( im, PX(i), PY(uvals[CLOSESET+j*3][i]),
                                             PX(i), PY(uvals[LOWSET+j*3][i]),
                                             _2D(ExtColor,LOWSET+j*3,i) );
                        if( uvals[HIGHSET+j*3][i] != GDC_NOVALUE )
                            gdImageLine( im, PX(i), PY(uvals[HIGHSET+j*3][i]),
                                             PX(i), PY(uvals[CLOSESET+j*3][i]),
                                             _2D(ExtColor,HIGHSET+j*3,i) );

                        if( psGDC->HLC_style & GDC_HLC_I_CAP )
                            {
                            if( uvals[LOWSET+j*3][i] != GDC_NOVALUE )
                                gdImageLine( im, PX(i)-hlf_hlccapwdth, PY(uvals[LOWSET+j*3][i]),
                                                 PX(i)+hlf_hlccapwdth, PY(uvals[LOWSET+j*3][i]),
                                                 _2D(ExtColor,LOWSET+j*3,i) );
                            if( uvals[HIGHSET+j*3][i] != GDC_NOVALUE )
                                gdImageLine( im, PX(i)-hlf_hlccapwdth, PY(uvals[HIGHSET+j*3][i]),
                                                 PX(i)+hlf_hlccapwdth, PY(uvals[HIGHSET+j*3][i]),
                                                 _2D(ExtColor,HIGHSET+j*3,i) );
                            }
                        if( psGDC->HLC_style & GDC_HLC_DIAMOND )
                            {
                            gdPoint         cd[4];

                            cd[0].x = PX(i)-hlf_hlccapwdth; cd[0].y = PY(uvals[CLOSESET+j*3][i]);
                            cd[1].x = PX(i);    cd[1].y = PY(uvals[CLOSESET+j*3][i])+hlf_hlccapwdth;
                            cd[2].x = PX(i)+hlf_hlccapwdth; cd[2].y = PY(uvals[CLOSESET+j*3][i]);
                            cd[3].x = PX(i);    cd[3].y = PY(uvals[CLOSESET+j*3][i])-hlf_hlccapwdth;
                            gdImageFilledPolygon( im, cd, 4, _2D(ExtColor,CLOSESET+j*3,i) );
                            }
                        }
                for( i=1; i<psGDC->num_points; ++i )
                    if( uvals[CLOSESET+j*3][i-1] != GDC_NOVALUE && uvals[CLOSESET+j*3][i] != GDC_NOVALUE )
                        {
                        if( psGDC->HLC_style & GDC_HLC_CLOSE_CONNECTED )   /* line from prev close */
                                gdImageLine( im, PX(i-1), PY(uvals[CLOSESET+j*3][i-1]),
                                                 PX(i), PY(uvals[CLOSESET+j*3][i-1]),
                                                 _2D(ExtColor,CLOSESET+j*3,i) );
                        else    /* CLOSE_CONNECTED and CONNECTING are mutually exclusive */
                        if( psGDC->HLC_style & GDC_HLC_CONNECTING )        /* thin connecting line */
                            gdImageLine( im, PX(i-1), PY(uvals[CLOSESET+j*3][i-1]),
                                             PX(i), PY(uvals[CLOSESET+j*3][i]),
                                             _2D(ExtColor,CLOSESET+j*3,i) );
                        }
                }
            break;
        }
        setno = 0;

    /* ---------- scatter points  over all other plots ---------- */
    /* scatters, by their very nature, don't lend themselves to standard array of points */
    /* also, this affords the opportunity to include scatter points onto any type of chart */
    /* drawing of the scatter point should be an exposed function, so the user can */
    /*  use it to draw a legend, and/or add their own */
    if( psGDC->scatter )
        {
        gdPoint ct[3];
        /*int       scatter_clr[psGDC->num_scatter_pts]; */
        DEF_ARRAY(int, scatter_clr, psGDC->num_scatter_pts);
        ALLOC_ARRAY(int, scatter_clr, psGDC->num_scatter_pts);

        for( i=0; i<psGDC->num_scatter_pts; ++i )
            {
            int     hlf_scatterwdth = (int)( (float)(PX(2)-PX(1))
                                             * (((float)((psGDC->scatter+i)->width)/100.0)/2.0) );
            int scat_x = PX( (psGDC->scatter+i)->point + (do_bar?1:0) ),
                scat_y = PY( (psGDC->scatter+i)->val );

            if( (psGDC->scatter+i)->point >= psGDC->num_points ||             /* invalid point */
                (psGDC->scatter+i)->point <  0 )
                continue;
            scatter_clr[i] = clrallocate( im, (psGDC->scatter+i)->color );

            switch( (psGDC->scatter+i)->ind )
                {
                case GDC_SCATTER_CIRCLE:
                    {
                    long    uniq_clr = get_uniq_color( im );
                    int     s        = 0,
                            e        = 360,
                            fo       = 0;

                    if( !do_bar )
                        if( (psGDC->scatter+i)->point == 0 )
                            { s = 270; e = 270+180; fo = 1; }
                        else
                        if( (psGDC->scatter+i)->point == psGDC->num_points-1 )
                            { s = 90; e = 90+180; fo = -1; }
                    if( uniq_clr != -1L )                           // the safe way
                        {
                        int uc = gdImageColorAllocate( im, l2gdcal(uniq_clr) );
                        gdImageArc( im, scat_x, scat_y,
                                        hlf_scatterwdth*2, hlf_scatterwdth*2,
                                        s, e,
                                        uc );
                        if( fo )                                    // close off  semi-circle case
                            gdImageLine( im, scat_x, scat_y+hlf_scatterwdth,
                                             scat_x, scat_y-hlf_scatterwdth,
                                             uc );
                        gdImageFillToBorder( im, scat_x+fo, scat_y, uc, scatter_clr[i] );
                        gdImageArc( im, scat_x, scat_y,
                                        hlf_scatterwdth*2, hlf_scatterwdth*2,
                                        s, e,
                                        scatter_clr[i] );
                        if( fo )
                            gdImageLine( im, scat_x, scat_y+hlf_scatterwdth,
                                             scat_x, scat_y-hlf_scatterwdth,
                                             scatter_clr[i] );
                        gdImageColorDeallocate( im, uc );
                        }
                    else                                            // chance it
                        {
                        gdImageArc( im, scat_x, scat_y,
                                        hlf_scatterwdth*2, hlf_scatterwdth*2,
                                        s, e,
                                        scatter_clr[i] );
                        if( fo )
                            gdImageLine( im, scat_x, scat_y+hlf_scatterwdth,
                                             scat_x, scat_y-hlf_scatterwdth,
                                             scatter_clr[i] );
                        gdImageFillToBorder( im, scat_x+fo, scat_y,
                                                 scatter_clr[i], scatter_clr[i] );
                        }
                    }
                    break;
                case GDC_SCATTER_TRIANGLE_UP:
                    ct[0].x = scat_x;
                    ct[0].y = scat_y;
                    ct[1].x = scat_x - hlf_scatterwdth;
                    ct[1].y = scat_y + hlf_scatterwdth;;
                    ct[2].x = scat_x + hlf_scatterwdth;
                    ct[2].y = scat_y + hlf_scatterwdth;
                    if( !do_bar )
                        if( (psGDC->scatter+i)->point == 0 )
                            ct[1].x = scat_x;
                        else
                        if( (psGDC->scatter+i)->point == psGDC->num_points-1 )
                            ct[2].x = scat_x;
                    gdImageFilledPolygon( im, ct, 3, scatter_clr[i] );
                    break;
                case GDC_SCATTER_TRIANGLE_DOWN:
                    ct[0].x = scat_x;
                    ct[0].y = scat_y;
                    ct[1].x = scat_x - hlf_scatterwdth;
                    ct[1].y = scat_y - hlf_scatterwdth;;
                    ct[2].x = scat_x + hlf_scatterwdth;
                    ct[2].y = scat_y - hlf_scatterwdth;
                    if( !do_bar )
                        if( (psGDC->scatter+i)->point == 0 )
                            ct[1].x = scat_x;
                        else
                        if( (psGDC->scatter+i)->point == psGDC->num_points-1 )
                            ct[2].x = scat_x;
                    gdImageFilledPolygon( im, ct, 3, scatter_clr[i] );
                    break;
                }
            }
        FREE_ARRAY(scatter_clr);
        }


/* overlay with a value and an arrow (e.g., total daily change) */
#ifdef THUMB_VALS
    /* put thmbl and thumbval over vol and plot lines */
    if( thumbnail )
        {
        int     n, d, w;
        char    thmbl[32];
        char    nmrtr[3+1], dmntr[3+1], whole[8];

        char    *dbg;
        dbg = GDC_price_to_str( ABS(thumbval),&n,&d,&w );

        sprintf( nmrtr, "%d", n );
        sprintf( dmntr, "%d", d );
        sprintf( whole, "%d", w );

        gdImageString( im,
                       gdFontSmall,
                       graphwidth/2-strlen(thumblabel)*SFONTWDTH/2,
                       1,
                       thumblabel,
                       ThumbLblColor );
        if( w || n )
            {
            int     chgcolor  = thumbval>0.0? ThumbUColor: ThumbDColor;
            int     thmbvalwidth = SFONTWDTH                      + /* up/down arrow */
                                   (w?strlen(whole)*SFONTWDTH: 0) + /* whole        */
                                   (n?strlen(nmrtr)*TFONTWDTH     + /* numerator    */
                                      SFONTWDTH                   + /* /            */
                                      strlen(dmntr)*TFONTWDTH:      /* denominator  */
                                      0);                           /* no frac part */

            smallarrow( im, graphwidth/2-thmbvalwidth/2, SFONTHGT, thumbval>0.0, chgcolor );
            if( w )
                {
                gdImageString( im,
                               gdFontSmall,
                               (graphwidth/2-thmbvalwidth/2)+SFONTWDTH,
                               SFONTHGT+2,
                               whole,
                               chgcolor );
                }
            if( n )
                {
                gdImageString( im,
                               gdFontTiny,
                               (graphwidth/2-thmbvalwidth/2)   +    /* start */
                               SFONTWDTH                       +    /* arrow */
                               (w? strlen(whole)*SFONTWDTH: 0) +    /* whole */
                               2,
                               SFONTHGT+2-2,
                               nmrtr,
                               chgcolor );
                gdImageChar  ( im,
                               gdFontSmall,
                               (graphwidth/2-thmbvalwidth/2)  +     /* start */
                               SFONTWDTH                      +     /* arrow */
                               (w? strlen(whole)*SFONTWDTH: 0) +    /* whole */
                               strlen(nmrtr)*TFONTWDTH,             /* numerator */
                               SFONTHGT+2,
                               '/',
                               chgcolor );
                gdImageString( im,
                               gdFontTiny,
                               (graphwidth/2-thmbvalwidth/2)  +     /* start */
                               SFONTWDTH                      +     /* arrow */
                               (w? strlen(whole)*SFONTWDTH: 0) +    /* whole */
                               strlen(nmrtr)*TFONTWDTH        +     /* numerator */
                               SFONTWDTH - 3,                       /* /     */
                               SFONTHGT+2+4,
                               dmntr,
                               chgcolor );
                }
            }
        }       /* thumblabel, thumbval */
#endif

    /* box it off */
    /*  after plotting so the outline covers any plot lines */
    {
    if( psGDC->border == GDC_BORDER_ALL || (psGDC->border & GDC_BORDER_X))
        gdImageLine( im,
                     flip_axis? PX(lowest): PX(0),
                     flip_axis? PY(0):PY(lowest), 
                     flip_axis? PX(lowest):PX(psGDC->num_points-1+(do_bar?2:0)), 
                     flip_axis? PY(psGDC->num_points-1+(do_bar?2:0)):PY(lowest), LineColor );

    if( psGDC->border == GDC_BORDER_ALL || (psGDC->border & GDC_BORDER_TOP))
        {
        setno = psGDC->stack_type==GDC_STACK_DEPTH? psGDC->num_hlc_sets? psGDC->num_hlc_sets: psGDC->num_groups: 1;
        gdImageLine( im,          
                     flip_axis? PXH2(1, highest): PX(0),
                     flip_axis? PY(0): PY(highest),
                     flip_axis? PXH2(1, highest):PX(psGDC->num_points-1+(do_bar?2:0)), 
                     flip_axis? PY(psGDC->num_points-1+(do_bar?2:0)): PY(highest), LineColor );
        setno = 0;
        }
    }
    if( psGDC->border )
        {
        int x1, y1, x2, y2;

        x1 = flip_axis? PXH2(1, highest):PX(0);
        y1 = flip_axis? PY(0):PY(highest);
        x2 = flip_axis? PX(lowest):PX(psGDC->num_points-1+(do_bar?2:0));
        y2 = flip_axis? PY(psGDC->num_points-1+(do_bar?2:0)): PY(lowest);
        if( psGDC->border == GDC_BORDER_ALL || (psGDC->border & GDC_BORDER_Y) )
            gdImageLine( im, flip_axis? PX(lowest):x1, 
                             flip_axis? y1:PY(lowest), x1, y1, LineColor );

        setno = set_depth;
        if( psGDC->border == GDC_BORDER_ALL || (psGDC->border & GDC_BORDER_Y) || (psGDC->border & GDC_BORDER_TOP))
            gdImageLine( im, x1, y1, 
                         flip_axis? PXH2(1, highest):PX(0),
                         flip_axis? PY(0):PY(highest), LineColor );
        /* if( !psGDC->grid || do_vol || psGDC->thumbnail )  */              /* grid leaves right side Y open */
            {
            if( psGDC->border == GDC_BORDER_ALL || (psGDC->border&GDC_BORDER_X) || (psGDC->border&GDC_BORDER_Y2))
                gdImageLine( im, x2, y2, 
                             flip_axis? PX(lowest):PX(psGDC->num_points-1+(do_bar?2:0)),
                             flip_axis? PY(psGDC->num_points-1+(do_bar?2:0)): PY(lowest), LineColor );
            if( psGDC->border == GDC_BORDER_ALL || (psGDC->border&GDC_BORDER_Y2))
                gdImageLine( im, flip_axis? PX(lowest):PX(psGDC->num_points-1+(do_bar?2:0)), 
                             flip_axis? PY(psGDC->num_points-1+(do_bar?2:0)): PY(lowest),
                             flip_axis? PXH2(1, highest): PX(psGDC->num_points-1+(do_bar?2:0)), 
                             flip_axis? PY(psGDC->num_points-1+(do_bar?2:0)): PY(highest), LineColor );
            }
        setno = 0;
        }

    if( psGDC->_0Shelf && threeD &&                             /* front of 0 shelf */
        ( (!do_bar && ((lowest < 0.0 && highest > 0.0) || 
                       ((lowest == 0.0 || highest == 0.0) && !(psGDC->border&GDC_BORDER_X) ) ) ) ||
          (do_bar && ((lowest < bars_ybase && highest > bars_ybase) || 
                      ((lowest == bars_ybase || highest == bars_ybase) && !(psGDC->border&GDC_BORDER_X) ) ) ) ) )
        {
        int x2 = PX( psGDC->num_points-1+(do_bar?2:0) ),
            y2 = PY( do_bar?bars_ybase:0 );

        gdImageLine( im, PX(0), PY(do_bar?bars_ybase:0), x2, y2, LineColor );     /* front line */
        setno = set_depth;
        gdImageLine( im, x2, y2, PX(psGDC->num_points-1+(do_bar?2:0)), PY(do_bar?bars_ybase:0), LineColor );
        setno = 0;                                              /* set back to foremost */
        }

    if( psGDC->annotation )            /* front half of annotation line */
        {
        int     x1 = PX(psGDC->annotation->point+(do_bar?1:0)),
                y1 = PY(highest);
        int     x2;
                                                            /* front line */
        gdImageLine( im, x1, PY(lowest)+1, x1, y1, AnnoteColor );
        if( threeD )
            {                                               /* on back plane */
            setno = set_depth;
            x2 = PX(psGDC->annotation->point+(do_bar?1:0));
                                                            /* prspective line */
            gdImageLine( im, x1, y1, x2, PY(highest), AnnoteColor );
            }
        else                                                /* for 3D done with back line */
            {
            x2 = PX(psGDC->annotation->point+(do_bar?1:0));
            gdImageLine( im, x1, y1, x1, y1-2, AnnoteColor );
            }
        /* line-to and note */
        if( *(psGDC->annotation->note) )                       /* any note? */
            {
            if( psGDC->annotation->point >= (psGDC->num_points/2) )       /* note to the left */
                {
                gdImageLine( im, x2,              PY(highest)-2,
                                 x2-annote_hgt/2, PY(highest)-2-annote_hgt/2,
                                 AnnoteColor );
                GDCImageStringNL( im,
                                  &GDC_fontc[psGDC->annotation_font],
                                  x2-annote_hgt/2-1-annote_len - 1,
                                  PY(highest)-annote_hgt+1,
                                  psGDC->annotation->note,
                                  AnnoteColor,
                                  GDC_JUSTIFY_RIGHT );
                }
            else                                                /* note to right */
                {
                gdImageLine( im, x2,              PY(highest)-2,
                                 x2+annote_hgt/2, PY(highest)-2-annote_hgt/2,
                                 AnnoteColor );
                GDCImageStringNL( im,
                                  &GDC_fontc[psGDC->annotation_font],
                                  x2+annote_hgt/2+1 + 1,
                                  PY(highest)-annote_hgt+1,
                                  psGDC->annotation->note,
                                  AnnoteColor,
                                  GDC_JUSTIFY_LEFT );
                }
            }
        setno = 0;
        }


    /* usually psGDC->generate_img is used in conjunction with hard or hold options */
    if( psGDC->generate_img )
        {
        fflush(img_fptr);           /* clear anything buffered  */
        switch( psGDC->image_type ) /* Can determine the image type */
            {
#ifdef HAVE_PNG
            case GDC_PNG:   
              gdImagePng( im, img_fptr );
              break;
#endif
#ifdef HAVE_JPEG
            case GDC_JPEG:  
              gdImageJpeg( im, img_fptr, psGDC->jpeg_quality );
              break;
#endif
#ifdef HAVE_WBMP
            case GDC_WBMP:  
              gdImageWBMP( im, PlotColor, img_fptr );
              break;
#endif
            case GDC_GIF:
            default:
              gdImageGif( im, img_fptr);
            }
        }

    if( bg_img )
        gdImageDestroy(bg_img);
    if( psGDC->hold_img & GDC_EXPOSE_IMAGE )
        psGDC->image = (void*)im;
    else
        gdImageDestroy(im);

    FREE_ARRAY(ExtVolColor);
    FREE_ARRAY(ExtColor);
    FREE_ARRAY(ExtColorShd);
    if (LineStyles)
    {
        for( j=0; j<psGDC->num_sets; ++j )
        {
            if (LineStyles[j].Style) free(LineStyles[j].Style);
        }
        FREE_ARRAY(LineStyles);
    }

    return 0;
}


#ifdef __TODO__NOT_IMPLEMENTED_YET__
/**********************************************************************
 *                       GDC_out_graph()
 *
 * This function is provided for backward compatibility.
 * 
 * It provides the same interface to the library that was present
 * in version 0.94b of the library for older programs.
 *
 * New programs should use the GDCChart*() set of functions instead.
 **********************************************************************/

int 
GDC_out_graph( short        IMGWIDTH,       /* no check for a gif that's */
               short        IMGHEIGHT,      /* too small to fit needed   */
                                            /* info (labels, etc), could */
                                            /* core dump */
               FILE         *img_fptr,      /* open file pointer (gif out)*/
               GDC_CHART_T  type,
               int          num_points,     /* points along x axis (even */
                                            /* interval) all arrays*/
                                            /* dependant on this */
               char         *xlbl[],        /* array of xlabels */
               int          num_sets,
                            ... )
{
    GDCChartPtr psGDC;
    int         retValue;

    /*-----------------------------------------------------------------
     *  Alloc and Init a chart handle
     *----------------------------------------------------------------*/
    psGDC = GDCChartAlloc( GDC_3DBAR, 250, 200);

    /*-----------------------------------------------------------------
     *  Call the lib
     *----------------------------------------------------------------*/
    retValue = GDCChartGIF( psGDC,
                   stdout,    /* FILE*       open FILE pointer */
                   6,         /* int         number of points per data set */
                   t,         /* char*[]     array of X labels */
                   2,         /* int         number of data sets */
                   a,         /* float[]     data set 1 */
                   b );       /*  ...        data set n */

    /*-----------------------------------------------------------------
     *  Cleanup and exit.
     *----------------------------------------------------------------*/
    GDCChartDestroy(psGDC);

    return retValue;
}

#endif /* __TODO__NOT_IMPLEMENTED_YET__ */


/**********************************************************************
 *                       GDC_set_values()
 *
 * This function is provided for backward compatibility with V0.94b.
 * 
 * Set values the old way, using a variable arg. list
 *
 * The arrays of values passed to the function are not copied, we just keep
 * a reference to them.  So it will be up to the caller to release the
 * memory associated with them (if necessary), but ONLY AFTER the chart
 * has been generated.
 *
 * Expected params (...) for each chart type:
 *     GDC_LINE
 *     GDC_BAR
 *     GDC_3DBAR
 *     GDC_3DAREA
 *     GDC_AREA            float   vals[], ...
 *
 * multiple sets make sense for rest?
 *
 *     GDC_HILOCLOSE       float   high[],
 *                         float   low[],
 *                         float   close[]
 *     
 *     GDC_COMBO_LINE_BAR
 *     GDC_COMBO_LINE_LINE
 *     GDC_COMBO_LINE_AREA float   val[],
 *                         float   vol[]
 *     
 *     GDC_COMBO_HLC_BAR
 *     GDC_COMBO_HLC_LINE
 *     GDC_COMBO_HLC_AREA  float   high[],
 *                         float   low[],
 *                         float   close[],
 *                         float   vol[]
 *
 * __TODO__ : FLOATINGBAR not supported yet.
 *
 **********************************************************************/

int  GDC_set_values( GDCChartPtr psGDC,
                     int         num_points, /* points along x axis (even   */
                                             /* interval) all arrays        */
                                             /* dependant on this           */
                     char        *xlbl[],    /* array of xlabels            */
                     int         num_sets,
                                 ... )
{
    va_list     ap;
    int         i;

    psGDC->num_points   = num_points;
    psGDC->num_groups   = num_sets;
    psGDC->num_sets     = num_sets;
    psGDC->num_hlc_sets = ( psGDC->type == GDC_COMBO_HLC_BAR   ||
                            psGDC->type == GDC_COMBO_HLC_AREA  ||
                            psGDC->type == GDC_COMBO_HLC_LINE  ||
                            psGDC->type == GDC_3DCOMBO_HLC_BAR ||
                            psGDC->type == GDC_3DCOMBO_HLC_AREA||
                            psGDC->type == GDC_3DCOMBO_HLC_LINE||
                            psGDC->type == GDC_3DHILOCLOSE     ||
                            psGDC->type == GDC_HILOCLOSE )? num_sets: 0;
    psGDC->num_h2_sets =  ( psGDC->type == GDC_H2BAR   ||
                            psGDC->type == GDC_3DH2BAR )? num_sets: 0;
    psGDC->num_fbar_sets =( psGDC->type == GDC_FLOATINGBAR   ||
                            psGDC->type == GDC_3DFLOATINGBAR )? num_sets: 0;

    psGDC->xlbl         = xlbl;

    /*-----------------------------------------------------------------
     * psGDC->num_sets is adjusted to become the real number of sets of
     * values in uvals[].  psGDC->num_groups gives the original num_sets value.
     *----------------------------------------------------------------*/
    psGDC->num_sets = psGDC->num_hlc_sets?(psGDC->num_hlc_sets*3):
                      psGDC->num_h2_sets?(psGDC->num_h2_sets*2):
                      psGDC->num_fbar_sets?(psGDC->num_fbar_sets*2): num_sets;

    /*-----------------------------------------------------------------
     * Alloc an array for the pointers to the user provided values.
     * Since this array is not provided by the user, we'll have to 
     * free it in GDCChartDestroy().
     *----------------------------------------------------------------*/
    if (psGDC->uvals != NULL && psGDC->uvals_allocated)
        free(psGDC->uvals);
    psGDC->uvals = (float **)calloc(psGDC->num_sets, sizeof(float*));
    if (psGDC->uvals == NULL)
    {
        fprintf(stderr, "Memory Allocation failed\n");
        return 1;
    }
    psGDC->uvals_allocated = TRUE;

    /*-----------------------------------------------------------------
     * get args  va number of float arrays
     *----------------------------------------------------------------*/
    va_start( ap, num_sets );
    for( i=0; i<psGDC->num_sets; i++ )
    {
        psGDC->uvals[i] = va_arg(ap, float*);
    }

    if( psGDC->type == GDC_COMBO_HLC_BAR   || /* aka: combo */
        psGDC->type == GDC_COMBO_HLC_AREA  ||
        psGDC->type == GDC_COMBO_HLC_LINE  ||
        psGDC->type == GDC_COMBO_LINE_BAR  ||
        psGDC->type == GDC_COMBO_LINE_AREA ||
        psGDC->type == GDC_COMBO_LINE_LINE ||
        psGDC->type == GDC_3DCOMBO_HLC_BAR ||
        psGDC->type == GDC_3DCOMBO_HLC_AREA||
        psGDC->type == GDC_3DCOMBO_HLC_LINE||
        psGDC->type == GDC_3DCOMBO_LINE_BAR||
        psGDC->type == GDC_3DCOMBO_LINE_AREA||
        psGDC->type == GDC_3DCOMBO_LINE_LINE )
    {
        psGDC->uvol = va_arg(ap, float*);
    }
    else
    {
        psGDC->uvol = NULL;
    }

    va_end(ap);

    return 0;
}
