/* GDCHART 2  GDCHART.H 
 *
 ************************************************************************
 * GDCHART.H is THE header file that should be #included by programs 
 * that want to use the GDChart library.
 *
 ************************************************************************
 * This version of the GDChart library was based on GDCHART 0.94b
 * (12 Nov. 1998) by Bruce Verderaime (brv@fred.net)
 *
 * Here is the original copyright notice:
 *
 +-----------------------------------------------------------------------
 | GDChart is free for use in your applications and for chart generation.
 | YOU MAY NOT re-distribute or represent the code as your own.
 | Any re-distributions of the code MUST reference the author, and include
 | any and all original documentaion.
 +-----------------------------------------------------------------------
 *
 * $Log: gdchart.h,v $
 * Revision 1.9  2002/11/15 18:58:34  julien
 * implement Bar Labels
 *
 * Revision 1.8  2002/02/11 19:51:10  julien
 * Add Combo_Line_HLC and 3DCombo_Line_HLC type
 *
 * Revision 1.7  2002/01/14 17:35:50  julien
 * Added ImageType and JPEGQuality parameters to create GIF, PNG and JPEG
 *
 * Revision 1.6  2002/01/11 15:51:16  julien
 * Add Combo_HLC_Line and 3DCombo_HLC_Line type
 *
 * Revision 1.5  2001/05/31 15:59:27  daniel
 * Single slice pie: draw a plain circle with no line from center to border.
 *
 * Revision 1.4  2001/02/16 15:47:58  daniel
 * Added pie_requested_radius param.
 *
 * Revision 1.3  2000/11/20 19:58:14  daniel
 * Integrate changes between v0.94b and v0.10.1dev
 *
 */

#ifndef _GDCHART_H
#define _GDCHART_H

#include <math.h>
#include "gd.h"

/*---------------------------------------------------------------------
 * GDCHART_VERSION... useful for programs to return version info.
 *--------------------------------------------------------------------*/
#define GDCHART_VERSION "2.2.0 (20020111)"

/*=====================================================================
                         Portability stuff
 =====================================================================*/

#ifndef M_PI
#  define M_PI   3.141592653589793
#endif
#ifndef M_PI_2
#  define M_PI_2 M_PI*M_PI
#endif

#ifndef TRUE
#  define TRUE    1
#  define FALSE   0
#endif

#ifndef ABS
#  define ABS( x )            ( (x)<0.0? -(x): (x) )
#endif

#ifndef MAX
#  define MAX( x, y )         ( (x)>(y)?(x):(y) )
#endif

#ifndef MIN
#  define MIN( x, y )         ( (x)<(y)?(x):(y) ) 
#endif

/*---------------------------------------------------------------------
 * FLT_MAX is defined by the ANSI C standard but MAXFLOAT is not
 *--------------------------------------------------------------------*/
#include <limits.h>
#include <float.h>
#ifndef MAXFLOAT
#  define MAXFLOAT FLT_MAX
#endif
#ifndef MAXINT
#  define MAXINT   INT_MAX
#endif
#ifndef MAXSHORT
#  define MAXSHORT SHRT_MAX
#endif

/*=====================================================================
                     General stuff (formerly from gdc.h)
 =====================================================================*/

#define MAX_NOTE_LEN        19
#define GDC_NOVALUE         -MAXFLOAT
#define GDC_NULL            GDC_NOVALUE
#define GDC_INTERP_VALUE    (GDC_NOVALUE/2.0) /* must set GDC_interpolations */
#define GDC_INTERP          ((psGDC->interpolations=TRUE),GDC_INTERP_VALUE)

#define GDC_NOCOLOR         0x1000000L
#define GDC_DFLTCOLOR       0x2000000L
#define PVRED               0x00FF0000
#define PVGRN               0x0000FF00
#define PVBLU               0x000000FF
#define l2gdcal( c )        ((c)&PVRED)>>16 , ((c)&PVGRN)>>8 , ((c)&0x000000FF)
#define l2gdshd( c )        (((c)&PVRED)>>16)/2 , (((c)&PVGRN)>>8)/2 , (((c)&0x000000FF))/2

/*---------------------------------------------------------------------
 * GDC_RGB_T: Type to use for RGB color values.  It is guaranteed to 
 * always be a 32 bits unsigned integer
 *--------------------------------------------------------------------*/
#if UINT_MAX == 65535
    typedef unsigned long   GDC_RGB_T;
#else
    typedef unsigned int    GDC_RGB_T;
#endif

/*---------------------------------------------------------------------
 *   GDC_FONT_SIZE_T:    Fonts, ordered by size
 *--------------------------------------------------------------------*/
typedef enum GDC_font_size 
{ 
    GDC_pad     = 0,
    GDC_TINY    = 1,
    GDC_SMALL   = 2,
    GDC_MEDBOLD = 3,
    GDC_LARGE   = 4,
    GDC_GIANT   = 5,
    GDC_numfonts= 6 
} GDC_FONT_SIZE_T;     /* GDC[PIE]_fontc depends on this */


/*---------------------------------------------------------------------
 *   GDC_HOLD_IMAGE_T
 *--------------------------------------------------------------------*/
typedef enum {
             GDC_DESTROY_IMAGE = 0,   /* default */
             GDC_EXPOSE_IMAGE  = 1,   /* user must call GDC_destroy_image() */
             GDC_REUSE_IMAGE   = 2    /* i.e., paint on top of */
             } GDC_HOLD_IMAGE_T;      /* EXPOSE & REUSE */


/*---------------------------------------------------------------------
 *   GDC_image_type_t
 *--------------------------------------------------------------------*/
typedef enum {
             GDC_GIF = 0,
             GDC_JPEG = 1,
             GDC_PNG = 2,
             GDC_WBMP = 3     /* as of gd1.8.3 WBMP is black and white only. */
             } GDC_image_type_t;

/*=====================================================================
             Stuff related to XY charts (formerly from gdchart.h)
 =====================================================================*/

typedef enum {
             /* ***** XY Chart types ***** */
             GDC_LINE,
             GDC_AREA,
             GDC_BAR,
             GDC_FLOATINGBAR,
             GDC_HILOCLOSE,
             GDC_COMBO_LINE_BAR,            /* aka, VOL[ume] */
             GDC_COMBO_HLC_BAR,
             GDC_COMBO_LINE_AREA,
             GDC_COMBO_LINE_LINE,
             GDC_COMBO_HLC_AREA,
             GDC_COMBO_HLC_LINE,
             GDC_COMBO_LINE_HLC,
             GDC_3DHILOCLOSE,
             GDC_3DCOMBO_LINE_BAR,
             GDC_3DCOMBO_LINE_AREA,
             GDC_3DCOMBO_LINE_LINE,
             GDC_3DCOMBO_HLC_BAR,
             GDC_3DCOMBO_HLC_AREA,
             GDC_3DCOMBO_HLC_LINE,
             GDC_3DCOMBO_LINE_HLC,
             GDC_3DBAR,
             GDC_3DFLOATINGBAR,
             GDC_3DAREA,
             GDC_3DLINE,
             /* ***** PIE Chart types ***** */
             GDC_3DPIE,
             GDC_2DPIE,
             /* ***** HORIZONTAL BAR Chart types ***** */
             GDC_HBAR,
             GDC_3DHBAR,
             GDC_H2BAR,
             GDC_3DH2BAR
            } GDC_CHART_T;

typedef enum {
             GDC_STACK_DEPTH,               /* "behind" (even non-3D) */
             GDC_STACK_SUM,
             GDC_STACK_BESIDE,
             GDC_STACK_LAYER
             } GDC_STACK_T;                 /* applies only to num_lines > 1 */

typedef enum {
             GDC_HLC_DIAMOND         = 1,
             GDC_HLC_CLOSE_CONNECTED = 2,   /* can't be used w/ CONNECTING */
             GDC_HLC_CONNECTING      = 4,   /* can't be used w/ CLOSE_CONNECTED */
             GDC_HLC_I_CAP           = 8
             } GDC_HLC_STYLE_T;             /* can be OR'd */

                                            /* only 1 annotation allowed */
typedef struct
            {
            float           point;          /* 0 <= point < num_points */
            GDC_RGB_T       color;
            char            note[MAX_NOTE_LEN+1];   /* NLs ok here */
            } GDC_ANNOTATION_T;

typedef enum {
             GDC_SCATTER_TRIANGLE_DOWN,
             GDC_SCATTER_TRIANGLE_UP,
             GDC_SCATTER_CIRCLE
             } GDC_SCATTER_IND_T;

typedef struct
            {
            float               point;      /* 0 <= point < num_points */
            float               val;
            int                 width;      /* % (1-100) */
            GDC_RGB_T           color;
            GDC_SCATTER_IND_T   ind;
            } GDC_SCATTER_T;

typedef enum {
            GDC_TICK_LABELS = -2,                  /* only at labels */
            GDC_TICK_POINTS = -1,                  /* each num_points */
            GDC_TICK_NONE   = 0
            /* > GDC_TICK_NONE */                  /* points & inter-point */
            } GDC_TICK_T;

typedef enum {                        /* backward compatible w/ FALSE/TRUE */
            GDC_BORDER_NONE = 0,
            GDC_BORDER_ALL  = 1,      /* should be GDC_BORDER_Y|Y2|X|TOP */
            GDC_BORDER_X    = 2,
            GDC_BORDER_Y    = 4,
            GDC_BORDER_Y2   = 8,
            GDC_BORDER_TOP  = 16
            } GDC_BORDER_T;

typedef enum {
             GDCPIE_PCT_NONE,
             GDCPIE_PCT_ABOVE,      /* relative to label, if any */
             GDCPIE_PCT_BELOW,
             GDCPIE_PCT_RIGHT,
             GDCPIE_PCT_LEFT
             } GDCPIE_PCT_TYPE;

typedef enum {
             GDC_SCALE_LINEAR,
             GDC_SCALE_LOG
             } GDC_SCALE_T;

/****************************************************/
/********** USER CHART OPTIONS w/ defaults **********/
/****************************************************/
/*---------------------------------------------------------------------
 *  GDCChart: Structure used to hold info about the current chart.
 *            It should be allocated using GDCChartAlloc(), and 
 *            destroyed using GDCChartDestroy().
 *            (Replaces the use of global variables in previous version)
 *--------------------------------------------------------------------*/
typedef struct _gdchart_t
{
    /**** PRIVATE MEMBERS ******************************/
    /* This first set of options is maintained by the lib. and should
     * not be set directly.
     */
    GDC_CHART_T         type;
    int                 imgwidth;
    int                 imgheight;

    int                 num_points;     /* num values along X axis */
    int                 num_sets;       /* Num. of sets of values  */
    int                 num_groups;     /* Num.Sets before num_sets adjusted */
    int                 num_hlc_sets;   /* Num. of sets of HLC values: 0 or 1*/
    int                 num_h2_sets;    /* Num. of sets of H2BAR values */
    int                 num_fbar_sets;  /* Num. of sets of FLOATINGBAR vals */
    char                **xlbl;         /* array of X axis labels  */
    float               **uvals;        /* array of sets (arrays) of values */
                                        /* size = num_sets + num_hlc_sets*3 */
    char                uvals_allocated;/* TRUE if the array uvals is owned */
                                        /* by the lib. (and should be freed) */
    float               *uvol;          /* For area and bar charts */

    /**** PUBLIC MEMBERS ******************************/
    /* The rest of the options can be set directly by the user... 
     * at least for now... ;-)
     */

    /**** COMMON OPTIONS ********************************/
    GDC_image_type_t    image_type;     /* DEFAULTO( GDC_GIF ) */
    int                 jpeg_quality;   /* DEFAULTO( -1 )    0-95 */
    char                generate_img;   /* DEFAULTO( TRUE )*/

    GDC_HOLD_IMAGE_T    hold_img;       /* DEFAULTO( GDC_DESTROY_IMAGE )*/
    void                *image;         /* DEFAULTO( (void*)NULL )  in/out */

    /**** XY chart options *****************************/
    char                *ytitle;
    char                *xtitle;
    char                *ytitle2;       /* ostesibly: volume label */
    char                *title;         /* NLs ok here */
    char                *ltitle;        /* Left/Right titles for H2BAR types*/
    char                *rtitle;
    GDC_FONT_SIZE_T     title_size;     /* DEFAULTO( GDC_MEDBOLD )*/
    GDC_FONT_SIZE_T     ytitle_size;    /* DEFAULTO( GDC_MEDBOLD )*/
    GDC_FONT_SIZE_T     xtitle_size;    /* DEFAULTO( GDC_MEDBOLD )*/
    GDC_FONT_SIZE_T     yaxisfont_size; /* DEFAULTO( GDC_SMALL )*/
    GDC_FONT_SIZE_T     xaxisfont_size; /* DEFAULTO( GDC_SMALL )*/
    char                *ylabel_fmt;    /* DEFAULTO( NULL )   printf fmt'ing, e.g.: "%.2f" */
    char                *ylabel2_fmt;   /* DEFAULTO( NULL )   default: "%.0f" future: fractions */
    char                *xlabel_ctl;    /* DEFAULTO( NULL )   num_points[] TRUE,FALSE */
    short               xlabel_spacing; /* DEFAULTO( 5 )      pixels  MAXSHORT means force all */
    char                ylabel_density; /* DEFAULTO( 80 )     % */
    char                interpolations; /* DEFAULTO( FALSE )   GDC_INTERP_VALUE in data */
    float               requested_ymin; /* DEFAULTO( GDC_NOVALUE )*/
    float               requested_ymax; /* DEFAULTO( GDC_NOVALUE )*/
    float               requested_ymin2;/* DEFAULTO( GDC_NOVALUE )*/
    float               requested_ymax2;/* DEFAULTO( GDC_NOVALUE )*/
    float               requested_yinterval; /* DEFAULTO( GDC_NOVALUE )*/
    char                _0Shelf;         /* DEFAULTO( TRUE )  if applicable */
    GDC_TICK_T          grid;           /* DEFAULTO( GDC_TICK_LABELS )*/
    GDC_TICK_T          ticks;          /* DEFAULTO( GDC_TICK_LABELS )*/
    char                xaxis;          /* DEFAULTO( TRUE )*/
    char                yaxis;          /* DEFAULTO( TRUE )*/
    char                yaxis2;         /* DEFAULTO( TRUE )*/
    char                yval_style;     /* DEFAULTO( TRUE )*/
    GDC_SCALE_T         yscale_type;    /* DEFAULTO( GDC_SCALE_LINEAR ) */
    GDC_STACK_T         stack_type;     /* DEFAULTO( GDC_STACK_DEPTH )*/
    float               _3d_depth;      /* DEFAULTO( 5.0 )   % img size */
    int                 _3d_angle;      /* DEFAULTO( 45 )    1-89 */
    int                 bar_width;      /* DEFAULTO( 75 )    % (1-100) */
    GDC_HLC_STYLE_T     HLC_style;      /* DEFAULTO( GDC_HLC_CLOSE_CONNECTED)*/
    int                 HLC_cap_width;  /* DEFAULTO( 25 )    % (1-100) */
    GDC_ANNOTATION_T    *annotation;    /* DEFAULTO( (GDC_ANNOTATION_T*)NULL)*/
    GDC_FONT_SIZE_T     annotation_font;/* DEFAULTO( GDC_SMALL )*/
    int                 num_scatter_pts;/* DEFAULTO( 0 )*/
    GDC_SCATTER_T       *scatter;       /* DEFAULTO( (GDC_SCATTER_T*)NULL )*/
    char                thumbnail;      /* DEFAULTO( FALSE )*/
    char                *thumblabel;
    float               thumbval;       /* DEFAULTO( -MAXFLOAT )*/
    GDC_BORDER_T        border;         /* DEFAULTO( GDC_BORDER_ALL )*/
    GDC_RGB_T           BGColor;        /* DEFAULTO( 0x000000L ) black */
    GDC_RGB_T           GridColor;      /* DEFAULTO( 0xA0A0A0L ) gray */
    GDC_RGB_T           LineColor;      /* DEFAULTO( GDC_DFLTCOLOR )*/
    GDC_RGB_T           PlotColor;      /* DEFAULTO( GDC_DFLTCOLOR )*/
    GDC_RGB_T           VolColor;       /* DEFAULTO( 0xA0A0FFL ) lgtblue1 */
    GDC_RGB_T           TitleColor;     /* DEFAULTO( GDC_DFLTCOLOR ) "opposite" of BG */
    GDC_RGB_T           XTitleColor;    /* DEFAULTO( GDC_DFLTCOLOR )*/
    GDC_RGB_T           YTitleColor;    /* DEFAULTO( GDC_DFLTCOLOR )*/
    GDC_RGB_T           YTitle2Color;   /* DEFAULTO( GDC_DFLTCOLOR )*/
    GDC_RGB_T           XLabelColor;    /* DEFAULTO( GDC_DFLTCOLOR )*/
    GDC_RGB_T           YLabelColor;    /* DEFAULTO( GDC_DFLTCOLOR )*/
    GDC_RGB_T           YLabel2Color;   /* DEFAULTO( GDC_DFLTCOLOR )*/
                            /* supercedes VolColor  ulong_color[num_points] */
    GDC_RGB_T           *ExtVolColor;   /* DEFAULTO( (GDC_RGB_T *)NULL )*/
                            /* supercedes LineColor ulong_color[num_sets] */
    GDC_RGB_T           *SetColor;      /* DEFAULTO( (GDC_RGB_T *)NULL )*/
                            /* supercedes SetColor  ulong_color[num_sets][num_points] */
    GDC_RGB_T           *ExtColor;      /* DEFAULTO( (GDC_RGB_T *)NULL )*/
    char                **LineStyles;   /* DEFAULTTO( NULL ) */
    char                transparent_bg; /* DEFAULTO( FALSE )*/
    char                *BGImage;       /* DEFAULTO( (char*)NULL )*/
    char                bar_labels;     /* DEFAULTO( FALSE )*/
    char                *bar_label_fmt; /* DEFAULTO( %.0f )*/

    /* legends?  separate img? */
    /* auto-size fonts, based on image size? */

    /* ----- following options are for expert users only ----- */
                        /* for alignment of multiple charts */
                        /* USE WITH CAUTION! */
    char                hard_size;      /* DEFAULTO( FALSE )*/
    int                 hard_xorig;     /* DEFAULTO( 0 )*/    /* in/out */
    int                 hard_graphwidth;/* DEFAULTO( 0 )*/    /* in/out */
    int                 hard_yorig;     /* DEFAULTO( 0 )*/    /* in/out */
    int                 hard_grapheight;/* DEFAULTO( 0 )*/    /* in/out */

    /******************************************************/
    /**** OPTIONS SPECIFIC TO PIE CHARTS  w/ defaults *****/
    /******************************************************/
    GDC_RGB_T           pie_EdgeColor;     /* DEFAULTO( GDC_NOCOLOR ) */
    char                pie_other_threshold; /*  DEFAULTO( -1 ) */
    GDC_FONT_SIZE_T     pie_label_size;    /*  DEFAULTO( GDC_SMALL ) */
    int                 pie_label_dist;    /*  DEFAULTO( 1+8/2 )   */
                                       /* 1+GDC_fontc[GDCPIE_label_size].h/2 */
    char                pie_label_line;    /*  DEFAULTO( FALSE ) */  /* from label to slice */
    int                 *pie_explode;      /* DEFAULTO( (int*)NULL ) */ /* [num_points] */
                        /* [num_points] supercedes GDCPIE_PlotColor */
    GDC_RGB_T           *pie_Color;        /* DEFAULTO( (GDC_RGB_T*)NULL ) */
    char                *pie_missing;      /* DEFAULTO( NULL )  TRUE/FALSE */

    GDCPIE_PCT_TYPE     pie_percent_labels;/* DEFAULTO( GDCPIE_PCT_NONE ) */
    char                *pie_percent_fmt;  /* DEFAULTO("%.0f%%")  printf fmt'ing */
    float               pie_requested_radius; /* DEFAULTO( GDC_NOVALUE )*/

} GDCChart;

typedef GDCChart * GDCChartPtr;


#ifdef __cplusplus
extern "C" {
#endif

/*---------------------------------------------------------------------
 *  Function prototypes  
 * (New API, these functions work with both XY and PIE charts)
 *--------------------------------------------------------------------*/

GDCChartPtr GDCChartAlloc(GDC_CHART_T type, int imgwidth, int imgheight);
void        GDCChartDestroy(GDCChartPtr psGDC);
int         GDCChartGIF( GDCChartPtr psGDC, FILE *fpOut);

int  GDCChartSetValues( GDCChartPtr psGDC,
                        int      num_points, /* points along x axis (even   */
                                             /* interval) all arrays        */
                                             /* dependant on this           */
                        char     *xlbl[],    /* array of xlabels            */
                        int      num_sets,
                        float    *pafValues );

/*---------------------------------------------------------------------
 *  Function prototypes  
 *  (Old API, provided only for backward compatibility with V0.94b
 *   The function of the new API (above) should be used for new apps)
 *--------------------------------------------------------------------*/
void    GDC_destroy_image( void* );
long    get_uniq_color( gdImagePtr im );

int 
GDC_out_graph( short        IMGWIDTH,       /* no check for a gif that's */
               short        IMGHEIGHT,      /* too small to fit needed   */
                                            /* info (labels, etc), could */
                                            /* core dump */
               FILE         *img_fptr,      /* open file pointer (gif out)*/
               GDC_CHART_T  type,
               int          num_points,     /* points along x axis (even */
                                            /* interval) all arrays*/
                                            /* dependant on this */
               char         *xlbl[],        /* array of xlabels */
               int          num_sets,
                            ... );

int  GDC_set_values( GDCChartPtr psGDC,
                     int         num_points, /* points along x axis (even   */
                                             /* interval) all arrays        */
                                             /* dependant on this           */
                     char        *xlbl[],    /* array of xlabels            */
                     int         num_sets,
                                 ... );

#ifdef __TODO__
void GDC_pie_gif( short         width,
                  short         height,
                  FILE*,                        /* open file pointer, can be stdout */
                  GDC_CHART_T,
                  int           num_points,
                  char          *labels[],      /* slice labels */
                  float         data[] );

#endif

#ifdef __cplusplus
}
#endif


#endif /*!_GDCHART_H*/


