/**********************************************************************
 * $Id: owttags.h,v 1.2 1998/12/09 02:08:06 daniel Exp $
 *
 * Name:     owttags.h
 * Project:  OWT HTML manipulation library
 * Purpose:  Header file for the Tags manipulation functions
 * Author:   Daniel Morissette, danmo@videotron.ca
 *
 **********************************************************************
 * Copyright (c) 1998, Daniel Morissette
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************
 *
 * $Log: owttags.h,v $
 * Revision 1.2  1998/12/09 02:08:06  daniel
 * Converted to use portability library
 *
 * Revision 1.1  1998/12/04 19:15:16  daniel
 * Initial revision
 *
 *
 **********************************************************************/

#ifndef _OWTTAGS_H_INCLUDED_
#define _OWTTAGS_H_INCLUDED_

#ifdef __cplusplus
extern "C" {
#endif

#include "cpl_string.h"

/*=====================================================================
                        Data Types and Constants
 =====================================================================*/

#define OWTTAG_START     "<!--OWT:"
#define OWTTAG_END       "-->"
  
/*---------------------------------------------------------------------
 *                        (struct) OWTTagInfo
 *
 * This structure is used to store information about a OWT Tag when
 * a HTML document is parsed.
 *--------------------------------------------------------------------*/
struct _OWTTagInfo
{
    char    *pszName;
    char    *pszType;
    char    *pszOptions;
    int     nFirstLine; /* Index of the opening tag          */
    int     nNumLines;  /* Including the start/end tag lines */
};
typedef struct _OWTTagInfo OWTTagInfo;

/*=====================================================================
                            Functions (owttags.c)
 =====================================================================*/

char      **OWTPreprocessTags(char **papszDoc);
OWTTagInfo *OWTGetNextTagInfo(char **papszDoc, int nFirstLine, int nLastLine);
void        OWTFreeTagInfo(OWTTagInfo *psTagInfo);
void        OWTPrintTagInfo(OWTTagInfo *psTagInfo);


#ifdef __cplusplus
}
#endif

#endif /* _OWTTAGS_H_INCLUDED_ */
