/**********************************************************************
 * $Id: owtcom_cmpnt.h,v 1.2 1999/06/09 19:27:20 daniel Exp $
 *
 * Name:     owtcom_cmpnt.h
 * Project:  OWT HTML manipulation library - COM Interface
 * Purpose:  Header file for the component classes
 * Author:   Daniel Morissette, danmo@videotron.ca
 *
 **********************************************************************
 * Copyright (c) 1999, Daniel Morissette
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************
 *
 * $Log: owtcom_cmpnt.h,v $
 * Revision 1.2  1999/06/09 19:27:20  daniel
 * Added GetText() and GetChildText()
 *
 * Revision 1.1.1.1  1999/06/04 19:37:36  daniel
 * First kind of complete version
 *
 **********************************************************************/

#ifndef _OWTCHT_H_INCLUDED_
#define _OWTCHT_H_INCLUDED_

#include "owtht.h"

#include "Iface.h"

/*=====================================================================
                            Class Definitions
 =====================================================================*/

/*---------------------------------------------------------------------
 *                        (class) OWTCHtBlock
 *
 * COM wrapper class for the OWTHtBlock class
 *--------------------------------------------------------------------*/

class OWTCHtBlock : public CUnknown,
                    public IOWTCHtBlock,
                    public ISupportErrorInfo
{
  private:
    OWTHtBlock *m_pBlock;

    // Pointer to type information.
    ITypeInfo* m_pITypeInfo ;

  public:	
    // Creation
    static HRESULT CreateInstance(IUnknown* pUnknownOuter,
                                  CUnknown** ppNewComponent ) ;

  private:
    // Declare the delegating IUnknown.
    DECLARE_IUNKNOWN

    // IUnknown
    virtual HRESULT __stdcall NondelegatingQueryInterface(const IID& iid,
                                                          void** ppv) ;

    // IDispatch
    virtual HRESULT __stdcall GetTypeInfoCount(UINT* pCountTypeInfo) ;

    virtual HRESULT __stdcall GetTypeInfo(
                UINT iTypeInfo,
                LCID,              // Localization is not supported.
                ITypeInfo** ppITypeInfo) ;
	
    virtual HRESULT __stdcall GetIDsOfNames(
                const IID& iid,
                OLECHAR** arrayNames,
                UINT countNames,
                LCID,              // Localization is not supported.
                DISPID* arrayDispIDs) ;

    virtual HRESULT __stdcall Invoke(   
	        DISPID dispidMember,
                const IID& iid,
                LCID,              // Localization is not supported.
                WORD wFlags,
                DISPPARAMS* pDispParams,
                VARIANT* pvarResult,
                EXCEPINFO* pExcepInfo,
                UINT* pArgErr) ;

    // ISupportErrorInfo
    virtual HRESULT __stdcall InterfaceSupportsErrorInfo(const IID& riid)
        {
            return (riid == IID_IOWTCHtBlock) ? S_OK : S_FALSE ;
        }

    // IOWTHtBlock Interface
    virtual HRESULT  __stdcall Draw(VARIANT vFileHandle);
    virtual HRESULT  __stdcall SetText(VARIANT vText);
    virtual HRESULT  __stdcall SetChildText(BSTR bstrName, VARIANT vText);

    virtual HRESULT  __stdcall GetText(VARIANT *pvabstrTextOut);
    virtual HRESULT  __stdcall GetChildText(BSTR bstrName, 
                                            VARIANT *pvabstrTextOut);

    virtual HRESULT  __stdcall GetName(BSTR *pbstrNameOut);
    virtual HRESULT  __stdcall SetName(BSTR bstrName);

    // Initialization
    virtual HRESULT Init() ;

    // Helper function to raise exceptions
    HRESULT RaiseException(HRESULT hrToReturn, const char *msg);

    // Constructor
    OWTCHtBlock(IUnknown* pUnknownOuter) ;

    // Destructor
    ~OWTCHtBlock() ;

} ;

/*---------------------------------------------------------------------
 *                        (class) OWTCHtDoc
 *
 * COM wrapper class for the OWTHtDoc class
 *--------------------------------------------------------------------*/

class OWTCHtDoc : public CUnknown,
                  public IOWTCHtDoc,
                  public ISupportErrorInfo
{
  private:
    OWTHtDoc *m_pDoc;

    // Pointer to type information.
    ITypeInfo* m_pITypeInfo ;

  public:	
    // Creation
    static HRESULT CreateInstance(IUnknown* pUnknownOuter,
                                  CUnknown** ppNewComponent ) ;

  private:
    // Declare the delegating IUnknown.
    DECLARE_IUNKNOWN

    // IUnknown
    virtual HRESULT __stdcall NondelegatingQueryInterface(const IID& iid,
                                                          void** ppv) ;

    // IDispatch
    virtual HRESULT __stdcall GetTypeInfoCount(UINT* pCountTypeInfo) ;

    virtual HRESULT __stdcall GetTypeInfo(
                UINT iTypeInfo,
                LCID,              // Localization is not supported.
                ITypeInfo** ppITypeInfo) ;
	
    virtual HRESULT __stdcall GetIDsOfNames(
                const IID& iid,
                OLECHAR** arrayNames,
                UINT countNames,
                LCID,              // Localization is not supported.
                DISPID* arrayDispIDs) ;

    virtual HRESULT __stdcall Invoke(   
	        DISPID dispidMember,
                const IID& iid,
                LCID,              // Localization is not supported.
                WORD wFlags,
                DISPPARAMS* pDispParams,
                VARIANT* pvarResult,
                EXCEPINFO* pExcepInfo,
                UINT* pArgErr) ;

    // ISupportErrorInfo
    virtual HRESULT __stdcall InterfaceSupportsErrorInfo(const IID& riid)
        {
            return (riid == IID_IOWTCHtDoc) ? S_OK : S_FALSE ;
        }

    // IOWTHtBlock Interface
    virtual HRESULT  __stdcall Draw(VARIANT vFileHandle);
    virtual HRESULT  __stdcall SetText(VARIANT vText);
    virtual HRESULT  __stdcall SetChildText(BSTR bstrName, VARIANT vText);

    virtual HRESULT  __stdcall GetText(VARIANT *pvabstrTextOut);
    virtual HRESULT  __stdcall GetChildText(BSTR bstrName, 
                                            VARIANT *pvabstrTextOut);

    virtual HRESULT  __stdcall GetName(BSTR *pbstrNameOut);
    virtual HRESULT  __stdcall SetName(BSTR bstrName);

    virtual HRESULT  __stdcall ChildExists(BSTR bstrName, 
                                           VARIANT_BOOL *pvbRetVal );
    virtual HRESULT  __stdcall SetChild(BSTR bstrName, IUnknown *pIUnknown);

    // IOWTHtDoc Interface
    virtual HRESULT  __stdcall LoadFromFile(BSTR bstrFilename);
    virtual HRESULT  __stdcall SaveToFile(BSTR bstrFilename);
    virtual HRESULT  __stdcall SetContentType(BSTR bstrContentType);

  private:
    // Initialization
    virtual HRESULT Init() ;

    // Helper function to raise exceptions
    HRESULT RaiseException(HRESULT hrToReturn, const char *msg);

    // Constructor
    OWTCHtDoc(IUnknown* pUnknownOuter) ;

    // Destructor
    ~OWTCHtDoc() ;

} ;

/*---------------------------------------------------------------------
 *                        (class) OWTCHtTable
 *
 * COM wrapper class for the OWTHtTable class
 *--------------------------------------------------------------------*/

class OWTCHtTable : public CUnknown,
                    public IOWTCHtTable,
                    public ISupportErrorInfo
{
  private:
    OWTHtTable *m_pTable;

    // Pointer to type information.
    ITypeInfo* m_pITypeInfo ;

  public:	
    // Creation
    static HRESULT CreateInstance(IUnknown* pUnknownOuter,
                                  CUnknown** ppNewComponent ) ;

  private:
    // Declare the delegating IUnknown.
    DECLARE_IUNKNOWN

    // IUnknown
    virtual HRESULT __stdcall NondelegatingQueryInterface(const IID& iid,
                                                          void** ppv) ;

    // IDispatch
    virtual HRESULT __stdcall GetTypeInfoCount(UINT* pCountTypeInfo) ;

    virtual HRESULT __stdcall GetTypeInfo(
                UINT iTypeInfo,
                LCID,              // Localization is not supported.
                ITypeInfo** ppITypeInfo) ;
	
    virtual HRESULT __stdcall GetIDsOfNames(
                const IID& iid,
                OLECHAR** arrayNames,
                UINT countNames,
                LCID,              // Localization is not supported.
                DISPID* arrayDispIDs) ;

    virtual HRESULT __stdcall Invoke(   
	        DISPID dispidMember,
                const IID& iid,
                LCID,              // Localization is not supported.
                WORD wFlags,
                DISPPARAMS* pDispParams,
                VARIANT* pvarResult,
                EXCEPINFO* pExcepInfo,
                UINT* pArgErr) ;

    // ISupportErrorInfo
    virtual HRESULT __stdcall InterfaceSupportsErrorInfo(const IID& riid)
        {
            return (riid == IID_IOWTCHtTable) ? S_OK : S_FALSE ;
        }

    // IOWTHtBlock Interface
    virtual HRESULT  __stdcall Draw(VARIANT vFileHandle);
    virtual HRESULT  __stdcall SetText(VARIANT vText);

    virtual HRESULT  __stdcall GetName(BSTR *pbstrNameOut);
    virtual HRESULT  __stdcall SetName(BSTR bstrName);

    // IOWTHtTable Interface
    virtual HRESULT  __stdcall AddColumn(VARIANT vabstrData);
    virtual HRESULT  __stdcall AddRow(VARIANT vabstrData);

    virtual HRESULT  __stdcall SetCell(LONG nCol, LONG nRow, BSTR bstrData, 
                                       VARIANT vbForceCreate);
    virtual HRESULT  __stdcall SetCellRange(LONG nCol1, LONG nRow1, 
                                            LONG numCols, LONG numRows,
                                            VARIANT vabstrData, 
                                            VARIANT vbForceCreate);
//    virtual HRESULT  __stdcall SetColumn(LONG nCol, VARIANT vabstrData );
//    virtual HRESULT  __stdcall SetRow(LONG nRow,VARIANT vabstrData );

    virtual HRESULT  __stdcall SetWrapMaxCols(LONG numCols);
    virtual HRESULT  __stdcall SetBorderWidth(LONG nBorderWidth);
    virtual HRESULT  __stdcall SetReversedOrientation(VARIANT_BOOL bRevOrient);
    virtual HRESULT  __stdcall SetNumHeaderRows(LONG numHeaderRows);
    virtual HRESULT  __stdcall SetNumHeaderColumns(LONG numHeaderRows);

  private:
    // Initialization
    virtual HRESULT Init() ;

    // Helper function to raise exceptions
    HRESULT RaiseException(HRESULT hrToReturn, const char *msg);

    // Constructor
    OWTCHtTable(IUnknown* pUnknownOuter) ;

    // Destructor
    ~OWTCHtTable() ;

} ;

/*---------------------------------------------------------------------
 *                        (class) OWTCHtOutput
 *
 * Output device to be used by COM components written in VB in their
 * Draw() method.
 *--------------------------------------------------------------------*/

class OWTCHtOutput : public CUnknown,
                     public IOWTCHtOutput,
                     public ISupportErrorInfo
{
  private:
    FILE *m_fpOut;

    // Pointer to type information.
    ITypeInfo* m_pITypeInfo ;

  public:	
    // Creation
    static HRESULT CreateInstance(IUnknown* pUnknownOuter,
                                  CUnknown** ppNewComponent ) ;

  private:
    // Declare the delegating IUnknown.
    DECLARE_IUNKNOWN

    // IUnknown
    virtual HRESULT __stdcall NondelegatingQueryInterface(const IID& iid,
                                                          void** ppv) ;

    // IDispatch
    virtual HRESULT __stdcall GetTypeInfoCount(UINT* pCountTypeInfo) ;

    virtual HRESULT __stdcall GetTypeInfo(
                UINT iTypeInfo,
                LCID,              // Localization is not supported.
                ITypeInfo** ppITypeInfo) ;
	
    virtual HRESULT __stdcall GetIDsOfNames(
                const IID& iid,
                OLECHAR** arrayNames,
                UINT countNames,
                LCID,              // Localization is not supported.
                DISPID* arrayDispIDs) ;

    virtual HRESULT __stdcall Invoke(   
	        DISPID dispidMember,
                const IID& iid,
                LCID,              // Localization is not supported.
                WORD wFlags,
                DISPPARAMS* pDispParams,
                VARIANT* pvarResult,
                EXCEPINFO* pExcepInfo,
                UINT* pArgErr) ;

    // ISupportErrorInfo
    virtual HRESULT __stdcall InterfaceSupportsErrorInfo(const IID& riid)
        {
            return (riid == IID_IOWTCHtOutput) ? S_OK : S_FALSE ;
        }

    // IOWTHtBlock Interface
    virtual HRESULT  __stdcall SetFileHandle(LONG nFileHandle);
    virtual HRESULT  __stdcall Write(VARIANT vText);

  private:
    // Initialization
    virtual HRESULT Init() ;

    // Helper function to raise exceptions
    HRESULT RaiseException(HRESULT hrToReturn, const char *msg);

    // Constructor
    OWTCHtOutput(IUnknown* pUnknownOuter) ;

    // Destructor
    ~OWTCHtOutput() ;

} ;


#endif /* _OWTHT_H_INCLUDED_ */
